import logging
import os
from functools import reduce

import numpy as np
import pandas as pd
import keras
import tensorflow as tf
from keras.layers import LeakyReLU
from sklearn.utils import shuffle

from exoml.iatson.iatson_og_generator import IatsonOgModelGenerator
from exoml.ml.layers.dropout import AdaptiveStdDropout
from exoml.ml.layers.transformer_classifier import TransformerClassifier
from exoml.ml.model.base_model import CategoricalPredictionSetStats
from exoml.ml.model.imbalanced_binary_model import ImbalancedBinaryModel


class IATSON_og(ImbalancedBinaryModel):
    def __init__(self, class_ids, class_names, type_to_label, hyperparams, channels=2, name='IATSON_og',
                 mode="all") -> None:
        super().__init__(name,
                         [11, 9 * 15, 300, 75, 75, 75, 75, 75, 75],
                         class_ids,
                         type_to_label,
                         hyperparams)
        self.channels = channels
        self.mode = mode

    def _get_focus_flux_conv_branch(self, name, channels=None, additional_inputs=None):
        if not channels:
            channels = self.channels
        leaky_relu_alpha = 0.01
        # TODO use binning error as channel
        focus_flux_input = keras.Input(shape=(self.input_size[3], channels), name=name)
        focus_flux_input_mod = focus_flux_input
        #focus_flux_err_input = keras.Input(shape=(self.input_size[3], channels), name=name + "_err")
        #focus_flux_input_mod = keras.layers.Dropout(self.hyperparams.dropout_rate)(focus_flux_input)
        #focus_flux_err_input_mod = keras.layers.Dropout(self.hyperparams.dropout_rate)(focus_flux_err_input)
        if self.hyperparams.white_noise_std is not None:
            focus_flux_input_mod = keras.layers.GaussianNoise(stddev=self.hyperparams.white_noise_std)(focus_flux_input_mod)
            #focus_flux_err_input_mod = keras.layers.GaussianNoise(stddev=self.hyperparams.white_noise_std)(focus_flux_err_input_mod)
        #focus_flux_input_final = keras.layers.concatenate([focus_flux_input_mod, focus_flux_err_input_mod], axis=2)
        focus_flux_input_final = focus_flux_input_mod
        focus_flux_branch = focus_flux_input_final
        #focus_flux_branch = keras.layers.normalization.batch_normalization.BatchNormalization()(focus_flux_input_final)
        # (time, flux, detrended_flux1... detrended_flux5, flux_model, centroidx, centroidy, motionx, motiony, bck)
        # flux model by transit params and stellar params
        #focus_flux_branch = keras.layers.SpatialDropout1D(rate=self.hyperparams.spatial_dropout_rate)(focus_flux_branch)
        focus_flux_branch = keras.layers.Conv1D(filters=32, kernel_size=20, padding="same",
                                                activation=LeakyReLU(leaky_relu_alpha), use_bias=True)(
            focus_flux_branch)
        focus_flux_branch = keras.layers.MaxPooling1D(pool_size=5, strides=1)(focus_flux_branch)
        #focus_flux_branch = keras.layers.normalization.batch_normalization.BatchNormalization()(focus_flux_branch)
        focus_flux_branch = keras.layers.SpatialDropout1D(rate=self.hyperparams.spatial_dropout_rate)(focus_flux_branch)
        focus_flux_branch = keras.layers.Conv1D(filters=64, kernel_size=10, padding="same",
                                                activation=LeakyReLU(leaky_relu_alpha), use_bias=True)(
            focus_flux_branch)
        #focus_flux_branch = keras.layers.normalization.batch_normalization.BatchNormalization()(focus_flux_branch)
        focus_flux_branch = keras.layers.MaxPooling1D(pool_size=3, strides=1)(focus_flux_branch)
        focus_flux_branch = keras.layers.SpatialDropout1D(rate=self.hyperparams.spatial_dropout_rate)(focus_flux_branch)
        focus_flux_branch = keras.layers.Conv1D(filters=128, kernel_size=5, padding="same",
                                                activation=LeakyReLU(leaky_relu_alpha), use_bias=True)(
            focus_flux_branch)
        #focus_flux_branch = keras.layers.normalization.batch_normalization.BatchNormalization()(focus_flux_branch)
        focus_flux_branch = keras.layers.MaxPooling1D(pool_size=3, strides=1)(focus_flux_branch)
        if additional_inputs is not None:
            additional_inputs_branch = additional_inputs
            if self.hyperparams.numerical_white_noise_std is not None:
                additional_inputs_branch = keras.layers.GaussianNoise(
                    stddev=self.hyperparams.numerical_white_noise_std)(additional_inputs_branch)
            focus_flux_branch = keras.layers.Concatenate(axis=1)(
                [focus_flux_branch, keras.layers.Flatten()(additional_inputs_branch)])
        focus_flux_branch = keras.layers.Dense(64, kernel_regularizer=tf.keras.regularizers.L1L2(l1=self.hyperparams.l1_regularization, l2=self.hyperparams.l2_regularization),
                                               activation=LeakyReLU(leaky_relu_alpha))(focus_flux_branch)
        focus_flux_branch = AdaptiveStdDropout(rate=self.hyperparams.dropout_rate,
                                         max_rate=self.hyperparams.dropout_max_rate)(focus_flux_branch)
        focus_flux_branch = keras.layers.LayerNormalization()(focus_flux_branch)
        return focus_flux_input, focus_flux_branch

    def _get_focus_flux_transformer_branch(self, name, transformer_blocks=6, transformer_heads=6, channels=None, additional_inputs=None):
        leaky_relu_alpha = 0.01
        # TODO use binning error as channel
        if channels is None:
            channels = self.channels
        focus_flux_input = keras.Input(shape=(self.input_size[3], channels), name=name)
        focus_flux_input_mod = focus_flux_input
        if self.hyperparams.white_noise_std is not None:
            focus_flux_input_mod = keras.layers.GaussianNoise(stddev=self.hyperparams.white_noise_std)(focus_flux_input_mod)
        focus_flux_input_final = focus_flux_input_mod
        transformer_input_size = self.input_size[3]
        transformer_output_size = self.input_size[3]
        kernel_size = 1
        classes = 128
        flux_branch = TransformerClassifier(transformer_input_size=transformer_input_size, patch_size=kernel_size,
                              num_heads=transformer_heads, mlp_dim=transformer_output_size, hyperparams=self.hyperparams,
                              num_blocks=transformer_blocks, classes=classes)(focus_flux_input_final)
        #flux_branch = keras.layers.GlobalMaxPooling1D()(flux_branch)
        flux_branch = keras.layers.Flatten()(flux_branch)
        if additional_inputs is not None:
            additional_inputs_branch = additional_inputs
            if self.hyperparams.numerical_white_noise_std is not None:
                additional_inputs_branch = keras.layers.GaussianNoise(
                    stddev=self.hyperparams.numerical_white_noise_std)(additional_inputs_branch)
            flux_branch = keras.layers.Concatenate(axis=1)(
                [flux_branch, keras.layers.Flatten()(additional_inputs_branch)])
        flux_branch = keras.layers.Dense(50, kernel_regularizer=tf.keras.regularizers.L1L2(
            l1=self.hyperparams.l1_regularization, l2=self.hyperparams.l2_regularization),
                                         activation=LeakyReLU(leaky_relu_alpha))(flux_branch)
        flux_branch = AdaptiveStdDropout(rate=self.hyperparams.dropout_rate,
                                         max_rate=self.hyperparams.dropout_max_rate)(flux_branch)
        flux_branch = keras.layers.LayerNormalization()(flux_branch)
        return focus_flux_input, flux_branch

    def build(self, use_transformers=False, transformer_blocks=6, transformer_heads=6):
        inputs, final_branch = self.build_transformer(transformer_blocks, transformer_heads) \
            if use_transformers else self.build_convolutional()
        self.set_model(keras.Model(inputs=inputs, outputs=final_branch, name=self.name))
        return self

    def build_convolutional(self):
        leaky_relu_alpha = 0.01
        og_input, og_branch = self._get_focus_flux_conv_branch("og_branch", 2)
        og_branch = keras.layers.Flatten()(og_branch)
        # og_branch = keras.layers.Dense(1000, kernel_regularizer=tf.keras.regularizers.L1L2(
        #     l1=self.hyperparams.l1_regularization, l2=self.hyperparams.l2_regularization),
        #                                activation=LeakyReLU(leaky_relu_alpha), name="final-dense2")(og_branch)
        # og_branch = AdaptiveStdDropout(rate=self.hyperparams.dropout_rate,
        #                                max_rate=self.hyperparams.dropout_max_rate, name="final-dropout2")(og_branch)
        # og_branch = keras.layers.Dense(250, kernel_regularizer=tf.keras.regularizers.L1L2(
        #     l1=self.hyperparams.l1_regularization, l2=self.hyperparams.l2_regularization),
        #                                activation=LeakyReLU(leaky_relu_alpha), name="final-dense3")(og_branch)
        # og_branch = AdaptiveStdDropout(rate=self.hyperparams.dropout_rate,
        #                                max_rate=self.hyperparams.dropout_max_rate, name="final-dropout3")(og_branch)
        # og_branch = keras.layers.Dense(32, kernel_regularizer=tf.keras.regularizers.L1L2(
        #     l1=self.hyperparams.l1_regularization, l2=self.hyperparams.l2_regularization),
        #                                activation=LeakyReLU(leaky_relu_alpha), name="final-dense4")(og_branch)
        # og_branch = AdaptiveStdDropout(rate=self.hyperparams.dropout_rate,
        #                                max_rate=self.hyperparams.dropout_max_rate, name="final-dropout4")(og_branch)
        og_branch = keras.layers.Dense(16, kernel_regularizer=tf.keras.regularizers.L1L2(
            l1=self.hyperparams.l1_regularization, l2=self.hyperparams.l2_regularization),
                                       activation=LeakyReLU(leaky_relu_alpha), name="final-dense5")(og_branch)
        og_branch = AdaptiveStdDropout(rate=self.hyperparams.dropout_rate,
                                       max_rate=self.hyperparams.dropout_max_rate, name="final-dropout5")(og_branch)
        og_branch = keras.layers.Dense(1, activation="sigmoid", name="logits")(og_branch)
        return og_input, og_branch

    def build_transformer(self, transformer_blocks=6, transformer_heads=6):
        leaky_relu_alpha = 0.01
        og_input, og_branch = self._get_focus_flux_transformer_branch("og_branch", transformer_blocks, transformer_heads, channels=2)
        og_branch = keras.layers.Dense(16, kernel_regularizer=tf.keras.regularizers.L1L2(
            l1=self.hyperparams.l1_regularization, l2=self.hyperparams.l2_regularization),
                                       activation=LeakyReLU(leaky_relu_alpha), name="final-dense6")(og_branch)
        og_branch = AdaptiveStdDropout(rate=self.hyperparams.dropout_rate,
                                       max_rate=self.hyperparams.dropout_max_rate, name="final-dropout6")(og_branch)
        final_branch = keras.layers.Dense(1, activation="sigmoid", name="logits")(og_branch)
        return og_input, final_branch

    def load_training_set(self, **kwargs):
        training_dir = kwargs.get('training_dir')
        injected_objects_df_q1_q17 = pd.read_csv(training_dir + "/q1_q17/injected_objects_tces_multi.csv", index_col=None)
        injected_objects_df_q1_q17 = injected_objects_df_q1_q17.loc[(injected_objects_df_q1_q17['multitype'].str.contains('tce')) |
                                                                    (injected_objects_df_q1_q17['type'] == 'planet') |
                                                                    (injected_objects_df_q1_q17['type'] == 'planet_transit')]
        df = [injected_objects_df_q1_q17]
        injected_objects_df = pd.concat(df)
        injected_objects_df.reset_index(inplace=True, drop=True)
        return injected_objects_df

    def instance_generator(self, dataset, dir, batch_size, input_sizes, type_to_label, zero_epsilon, shuffle=True):
        return IatsonOgModelGenerator(dataset, dir, batch_size, input_sizes, type_to_label, zero_epsilon, shuffle_batch=shuffle)

    def predict(self, input):
        prediction = self.model(input, training=False)
        return self.predict_threshold([prediction])

    def predict_batch(self, inputs, expected_outputs=None, dataset=None, training_dir=None, plot_mismatches=False,
                      batch_size=20, cores=os.cpu_count() - 1, threshold=[0.5]):
        predictions = self.model.predict(inputs, use_multiprocessing=True, batch_size=batch_size, workers=cores)
        max_prediction_indexes, max_prediction_values = self.predict_threshold(predictions, threshold)
        if expected_outputs is not None:
            planet_stats = self.test_metrics(expected_outputs, max_prediction_indexes, max_prediction_values, predictions,
                                             dataset, training_dir=training_dir, plot_mismatches=plot_mismatches)
            return max_prediction_indexes, max_prediction_values, [planet_stats]
        else:
            return max_prediction_indexes, max_prediction_values

    def predict_threshold(self, predictions, threshold=[0.5]):
        max_prediction_indexes = np.array([np.argmax(prediction) for prediction in predictions])
        max_prediction_values = np.array([predictions[index][max_prediction_index]
                                 if predictions[index][max_prediction_index] > threshold[max_prediction_index]
                                 else -1
                                 for index, max_prediction_index in enumerate(max_prediction_indexes)])
        max_prediction_indexes[np.argwhere(max_prediction_values < 0).flatten()] = -1
        return max_prediction_indexes, max_prediction_values

    def predict_df(self, df, training_dir, batch_size=20, cores=os.cpu_count() - 1, zero_epsilon=1e-7,
                   thresholds=[0.5], plot_mismatches=False):
        expected_outputs = None
        testing_batch_generator = IatsonOgModelGenerator(df, training_dir, batch_size,
                                                             self.input_size,
                                                             self.type_to_label, zero_epsilon, from_arrays=True,
                                                             shuffle_batch=False)
        for i in np.arange(0, len(df) // batch_size + 1, 1):
            input, expected_output = testing_batch_generator.__getitem__(i)
            expected_outputs = np.concatenate((expected_outputs, expected_output)) if expected_outputs is not None else expected_output
        for threshold in thresholds:
            max_prediction_indexes, max_prediction_values, prediction_stats = \
                self.predict_batch(testing_batch_generator, expected_outputs, df, training_dir, plot_mismatches,
                                   batch_size, cores=cores, threshold=[threshold])
            prediction_stats = prediction_stats[0]
            logging.info("Prediction stats for label %.0f and threshold %s: TP=%.0f, FP=%.0f, FN=%.0f, ACC=%.3f, PRE=%.3f, REC=%.3f", 0,
                         threshold, prediction_stats.tp, prediction_stats.fp, prediction_stats.fn, prediction_stats.accuracy,
                         prediction_stats.precision, prediction_stats.recall)
            logging.info("Prediction stats @ k top predictions: \n%s", prediction_stats.k_df.to_string())
            logging.info("Mismatches: \n%s", prediction_stats.predictions_df.to_string())
            logging.info("Mismatches by types: \n%s", prediction_stats.predictions_df
                         .groupby(by=['type', 'predicted_class'])['object_id'].count().to_string())

    def predict_test_set(self, training_dir, model_dir, batch_size=20, cores=os.cpu_count() - 1, test_set_limit=None,
                         zero_epsilon=1e-7, plot_mismatches=False):
        test_dataset = pd.read_csv(model_dir + '/test_dataset.csv')
        if test_set_limit is not None:
            test_dataset = test_dataset.iloc[0:test_set_limit]
        expected_outputs = None
        testing_batch_generator = IatsonOgModelGenerator(test_dataset, training_dir, batch_size,
                                                             self.input_size,
                                                             self.type_to_label, zero_epsilon, from_arrays=True,
                                                             shuffle_batch=False)
        for i in np.arange(0, len(test_dataset) // batch_size + 1, 1):
            input, expected_output = testing_batch_generator.__getitem__(i)
            expected_outputs = np.concatenate((expected_outputs, expected_output)) if expected_outputs is not None else expected_output
        for threshold in np.arange(0.5, 1, 0.01):
            max_prediction_indexes, max_prediction_values, prediction_stats = \
                self.predict_batch(testing_batch_generator, expected_outputs, test_dataset, training_dir, plot_mismatches,
                                   batch_size, cores=cores, threshold=[threshold])
            prediction_stats = prediction_stats[0]
            logging.info("Prediction stats for label %.0f and threshold %s: TP=%.0f, FP=%.0f, FN=%.0f, ACC=%.3f, PRE=%.3f, REC=%.3f", 0,
                         threshold, prediction_stats.tp, prediction_stats.fp, prediction_stats.fn, prediction_stats.accuracy,
                         prediction_stats.precision, prediction_stats.recall)
            logging.info("Prediction stats @ k top predictions: \n%s", prediction_stats.k_df.to_string())
            logging.info("Mismatches: \n%s", prediction_stats.predictions_df.to_string())
            logging.info("Mismatches by types: \n%s", prediction_stats.predictions_df
                         .groupby(by=['type', 'predicted_class'])['object_id'].count().to_string())


    def test_metrics(self, expected_outputs, max_prediction_indexes, max_prediction_values, predictions,
                     dataset=None, training_dir=None, plot_mismatches=False):
        expected_outputs = expected_outputs.flatten()
        positive_preds = np.argwhere(max_prediction_indexes.flatten() != -1).flatten()
        negative_preds = np.argwhere(max_prediction_indexes.flatten() == -1).flatten()
        positive_outputs = np.argwhere(expected_outputs == 1).flatten()
        negative_outputs = np.argwhere(expected_outputs == 0).flatten()
        label_tp_indexes = np.intersect1d(positive_preds, positive_outputs)
        label_tn_indexes = np.intersect1d(negative_preds, negative_preds)
        label_fp_indexes = positive_preds[np.isin(positive_preds, negative_outputs)]
        label_fn_indexes = negative_preds[np.isin(negative_preds, positive_outputs)]
        label_tp = len(label_tp_indexes)
        label_tn = len(label_tn_indexes)
        label_fp = len(label_fp_indexes)
        label_fn = len(label_fn_indexes)
        accuracy = (label_tp + label_tn) / len(expected_outputs)
        precision = label_tp / (label_tp + label_fp + 1e-7)
        recall = label_tp / (label_tp + label_fn + 1e-7)
        max_predictions = np.array([np.max(prediction) for prediction in predictions])
        max_predictions_sort_args = np.flip(np.argsort(max_predictions))
        max_predictions_sort = max_predictions[max_predictions_sort_args]
        max_prediction_indexes_sort = max_prediction_indexes[max_predictions_sort_args]
        max_prediction_values_sort = max_prediction_values[max_predictions_sort_args]
        expected_outputs_sort = expected_outputs[max_predictions_sort_args]
        k_df = pd.DataFrame(columns=['k', 'precision', 'recall', 'accuracy', 'tp', 'fp', 'tn', 'fn'])
        if 100 < len(expected_outputs):
            for k in np.arange(100, len(expected_outputs) if len(expected_outputs) < 2500 else 2500, 100):
                max_predictions_sort_k = max_predictions_sort[0:k]
                max_prediction_indexes_sort_k = max_prediction_indexes_sort[0:k]
                max_prediction_values_sort_k = max_prediction_values_sort[0:k]
                expected_outputs_sort_k = expected_outputs_sort[0:k]
                positive_preds = np.argwhere(max_prediction_indexes_sort_k.flatten() != -1).flatten()
                negative_preds = np.argwhere(max_prediction_indexes_sort_k.flatten() == -1).flatten()
                positive_outputs = np.argwhere(expected_outputs_sort_k == 1).flatten()
                negative_outputs = np.argwhere(expected_outputs_sort_k == 0).flatten()
                label_tp_indexes = np.intersect1d(positive_preds, positive_outputs)
                label_tn_indexes = np.intersect1d(negative_preds, negative_preds)
                label_fp_indexes = positive_preds[np.isin(positive_preds, negative_outputs)]
                label_fn_indexes = negative_preds[np.isin(negative_preds, positive_outputs)]
                label_tp = len(label_tp_indexes)
                label_tn = len(label_tn_indexes)
                label_fp = len(label_fp_indexes)
                label_fn = len(label_fn_indexes)
                accuracy = (label_tp + label_tn) / len(expected_outputs_sort_k)
                precision = label_tp / (label_tp + label_fp + 1e-7)
                recall = label_tp / (label_tp + label_fn + 1e-7)
                k_df = k_df.append(
                    {'k': k, 'precision': precision, 'recall': recall, 'accuracy': accuracy, 'tp': label_tp,
                     'fp': label_fp, 'fp': label_fp, 'fn': label_fn}, ignore_index=True)
        mismatches_df = pd.DataFrame(columns=['object_id', 'type', 'expected_class', 'predicted_class',
                                              'prediction_value'])
        if dataset is not None:
            found_classes = [0 if prediction  == -1 else 1 for prediction in max_prediction_values]
            dataset = dataset.reset_index(drop=True)
            for i, row in dataset.iterrows():
                object_id = row['object_id']
                period = row['period']
                id = object_id + "_" + str(period)
                expected_tag = row['type']
                expected_output = expected_outputs[i]
                predicted_output = found_classes[i]
                predicted_value = predictions[i][max_prediction_indexes[i]]
                if expected_output != predicted_output:
                    mismatches_df = mismatches_df.append({'object_id': object_id + '_' + str(round(period, 2)), "type": expected_tag,
                                          "expected_class": expected_output, "predicted_class": predicted_output,
                                          "prediction_value": predicted_value}, ignore_index=True)
                    if plot_mismatches:
                        logging.info("%s with label %s mismatched with value %s", id, expected_tag, predicted_value)
                        IatsonOgModelGenerator(dataset[(dataset['object_id'] == object_id) & (dataset['period'] == period)],
                                                                             training_dir, 1,
                                                                             self.input_size,
                                                                             self.type_to_label,
                                                                             from_arrays=True,
                                                                             shuffle_batch=False,
                                                                             plot_inputs=True).__getitem__(0)
        mismatches_df = mismatches_df.sort_values(by=['type', 'predicted_class'])
        return CategoricalPredictionSetStats(label_tp, label_fp, label_fn, accuracy, precision, recall, k_df,
                                             mismatches_df)



import logging

from keras.losses import CategoricalCrossentropy
from keras.metrics import CategoricalAccuracy, Precision, Recall, AUC, RecallAtPrecision
from tensorflow_addons.metrics import F1Score
import pandas as pd
from exoml.ml.model.base_model import BaseModel
import matplotlib.pyplot as plt
from matplotlib.pyplot import cm
import numpy as np


class CategoricalModel(BaseModel):
    def __init__(self, class_names, name, input_size, class_ids, dropout_rate=0.1) -> None:
        super().__init__(name, input_size, class_ids, dropout_rate)
        self.class_names = class_names

    def instance_loss_accuracy(self):
        return CategoricalCrossentropy(), CategoricalAccuracy()

    def instance_metrics(self):
        metrics = []
        for index, class_name in enumerate(self.class_names):
            class_name_string = str(class_name)
            metrics = metrics + [Precision(class_id=index, name=class_name_string + "_precision", thresholds=0.5),
                                 Recall(class_id=index, name=class_name_string + "_recall", thresholds=0.5),
                                 RecallAtPrecision(0.99, class_id=index, name=class_name_string + "_r@p99"),
                                 RecallAtPrecision(0.975, class_id=index, name=class_name_string + "_r@p975"),
                                 RecallAtPrecision(0.95, class_id=index, name=class_name_string + "_r@p95")]
        classes_count = len(self.class_names)
        return metrics + [F1Score(num_classes=classes_count, average='macro'),
                          AUC(multi_label=True, num_labels=classes_count, name="roc_auc"),
                          AUC(multi_label=True, curve="PR", num_labels=classes_count, name="pr_auc")]

    def plot_metrics(self, model_dir, steps_per_epoch, with_validation=False):
        logging.info("Plotting metrics")
        training_log_df = pd.read_csv(model_dir + '/training_log.csv')
        training_log_df['batch_total'] = training_log_df['batch'] + (training_log_df['epoch'] * steps_per_epoch)
        validation_log_df = training_log_df.dropna()
        fig, axs = plt.subplots(3, 2, figsize=(25, 12), constrained_layout=True)
        fig.suptitle(self.name + ' Model Metrics', fontsize=14)
        axs[0][0].plot(training_log_df['batch_total'], training_log_df['lr'])
        axs[0][0].set_title("Learning rate")
        axs[0][0].set_xlabel("Batch")
        secx = axs[0][0].secondary_xaxis('top',
                                         functions=(lambda x: x / steps_per_epoch, lambda x: x / steps_per_epoch))
        secx.set_xlabel('Epoch')
        axs[0][0].set_ylabel("Learning rate")
        axs[0][0].legend(loc='upper right')
        axs[0][1].plot(training_log_df['batch_total'], training_log_df['loss'], color='blue', label='Train loss')
        axs[2][1].plot(training_log_df['batch_total'], training_log_df['categorical_accuracy'], color='black',
                       label='Train Acc.')
        axs[2][1].plot(training_log_df['batch_total'], training_log_df['f1_score'],
                       color='red', label='Train F1')
        axs[2][1].plot(training_log_df['batch_total'], training_log_df['roc_auc'],
                       color='blue', label='Train ROC AUC')
        axs[2][1].plot(training_log_df['batch_total'], training_log_df['pr_auc'],
                       color='green', label='Train PR AUC')
        if with_validation:
            validation_log_df['batch_total'] = (validation_log_df['epoch'] + 1) * (steps_per_epoch - 1)
            axs[0][1].plot(validation_log_df['batch_total'], validation_log_df['val_loss'],
                           color='blue', label='Val. loss', linestyle='--')
            axs[2][1].plot(validation_log_df['batch_total'],
                           validation_log_df['val_categorical_accuracy'], color='black', label='Val. acc.',
                           linestyle='--')
            axs[2][1].plot(validation_log_df['batch_total'],
                           validation_log_df['val_f1_score'], color='red', label='Val. F1', linestyle='--')
            axs[2][1].plot(validation_log_df['batch_total'],
                           validation_log_df['val_roc_auc'], color='blue', label='Val. ROC AUC', linestyle='--')
            axs[2][1].plot(validation_log_df['batch_total'],
                           validation_log_df['val_pr_auc'], color='green', label='Val. PR AUC', linestyle='--')
        axs[0][1].set_title("Loss")
        axs[0][1].set_xlabel("Batch")
        secx = axs[0][1].secondary_xaxis('top',
                                         functions=(lambda x: x / steps_per_epoch, lambda x: x / steps_per_epoch))
        secx.set_xlabel('Epoch')
        axs[0][1].set_ylabel("Loss")
        axs[0][1].legend(loc='upper right')
        axs[2][1].set_title("Accuracy, F1 Score, ROC and PR AUCs")
        axs[2][1].set_xlabel("Batch")
        secx = axs[2][1].secondary_xaxis('top',
                                         functions=(lambda x: x / steps_per_epoch, lambda x: x / steps_per_epoch))
        secx.set_xlabel('Epoch')
        axs[2][1].set_ylabel("Accuracy/F1/AUCs")
        axs[2][1].legend(loc='upper left')
        classes_count = len(self.class_names)
        colors = cm.rainbow(np.linspace(0, 1, classes_count))
        for index, class_name in enumerate(self.class_names):
            class_name_string = str(class_name)
            axs[1][0].plot(training_log_df['batch_total'], training_log_df[class_name_string + '_precision'],
                           color=colors[index], label='Train ' + class_name_string + ' Prec.')
            axs[1][1].plot(training_log_df['batch_total'], training_log_df[class_name_string + '_recall'],
                           color=colors[index], label='Train ' + class_name_string + ' Recall.')
            axs[2][0].plot(training_log_df['batch_total'], training_log_df[class_name_string + '_r@p99'],
                           color=colors[index], label='Train ' + class_name_string + ' r@p.')
            if with_validation:
                axs[1][0].plot(validation_log_df['batch_total'],
                               validation_log_df['val_' + class_name_string + '_precision'],
                               color=colors[index], label='Val. ' + class_name_string + ' prec.', linestyle='--')
                axs[1][1].plot(validation_log_df['batch_total'],
                               validation_log_df['val_' + class_name_string + '_recall'],
                               color=colors[index], label='Val. ' + class_name_string + ' recall.', linestyle='--')
                axs[2][0].plot(validation_log_df['batch_total'],
                               validation_log_df['val_' + class_name_string + '_r@p99'],
                               color=colors[index], label='Val. ' + class_name_string + ' r@p', linestyle='--')
        axs[1][0].set_title("Precision (threshold = 0.5)")
        axs[1][0].set_xlabel("Batch")
        secx = axs[1][0].secondary_xaxis('top',
                                         functions=(lambda x: x / steps_per_epoch, lambda x: x / steps_per_epoch))
        secx.set_xlabel('Epoch')
        axs[1][0].set_ylabel("Precision")
        axs[1][0].legend(loc='upper left')
        axs[1][1].set_title("Recall (threshold = 0.5)")
        axs[1][1].set_xlabel("Batch")
        secx = axs[1][1].secondary_xaxis('top',
                                         functions=(lambda x: x / steps_per_epoch, lambda x: x / steps_per_epoch))
        secx.set_xlabel('Epoch')
        axs[1][1].set_ylabel("Recall")
        axs[1][1].legend(loc='upper left')
        axs[2][0].set_title("Recall at Precision=0.99")
        axs[2][0].set_xlabel("Batch")
        secx = axs[2][0].secondary_xaxis('top',
                                         functions=(lambda x: x / steps_per_epoch, lambda x: x / steps_per_epoch))
        secx.set_xlabel('Epoch')
        axs[2][0].set_ylabel("R@P")
        axs[2][0].legend(loc='upper left')
        plt.savefig(model_dir + '/metrics.png')
        plt.close(fig)
        plt.clf()
        logging.info("Plotted metrics")



import logging
from abc import ABC
import matplotlib.pyplot as plt
import random
from typing import Tuple
import ellc
import numpy as np
from numpy import ndarray
from numpy.random import default_rng
import astropy.units as u
import astropy.constants as ac
import timesynth
from sklearn.preprocessing import MinMaxScaler


class FeaturesInjector(ABC):
    random_number_generator = default_rng()

    def __init__(self, time_series: ndarray):
        super().__init__()
        self.time_series = time_series

    def inject(self) -> tuple[ndarray, ndarray, ndarray]:
        pass

class RedNoiseInjector(FeaturesInjector):
    def __init__(self, time_series: ndarray, power: tuple[float, float], period: tuple[float, float], number_of_signals: int):
        super().__init__(time_series)
        self.power = power
        self.period = period
        self.number_of_signals = number_of_signals

    def inject(self) -> tuple[ndarray, ndarray, ndarray]:
        number_of_red_noise = random.uniform(0, 100)
        if number_of_red_noise < 50:
            number_of_red_noise = 0
        elif number_of_red_noise < 75:
            number_of_red_noise = 1
        elif number_of_red_noise < 90:
            number_of_red_noise = 2
        else:
            number_of_red_noise = int(random.uniform(0, self.number_of_signals))
        flux: ndarray = np.ones(self.time_series.shape[1])
        for index in range(0, number_of_red_noise):
            red_noise_freq: float = self.random_number_generator.uniform(self.period[0], self.period[1])
            red_noise_power: float = self.random_number_generator.uniform(self.power[0], self.power[1])
            logging.info(f'Injecting red noise with power={red_noise_power} and freq={red_noise_freq}')
            red_noise = timesynth.noise.RedNoise(std=red_noise_power, tau=red_noise_freq)
            timeseries_rn: ndarray = np.zeros(self.time_series.shape[1])
            for time_index, value in enumerate(self.time_series[0]):
                rn_value = red_noise.sample_next(value, None, None)
                rn_value = rn_value[0] if isinstance(rn_value, (list, np.ndarray)) else rn_value
                timeseries_rn[time_index] = rn_value
            flux = flux + timeseries_rn
        return self.time_series[0], self.time_series[1] + flux - 1, flux


class SinusoidalInjector(FeaturesInjector):
    def __init__(self, time_series: ndarray, amplitude: tuple[float, float], period: tuple[float, float], number_of_signals: int):
        super().__init__(time_series)
        self.power = amplitude
        self.period = period
        self.number_of_signals = number_of_signals

    def inject(self) -> tuple[ndarray, ndarray, ndarray]:
        number_of_signals = random.uniform(0, 100)
        if number_of_signals < 50:
            number_of_signals = 0
        elif number_of_signals < 75:
            number_of_signals = 1
        elif number_of_signals < 90:
            number_of_signals = 2
        else:
            number_of_signals = int(random.uniform(0, self.number_of_signals))
        flux: ndarray = np.ones(self.time_series.shape[1])
        for index in range(0, number_of_signals):
            period: float = self.random_number_generator.uniform(self.period[0], self.period[1])
            amplitude: float = self.random_number_generator.uniform(self.power[0], self.power[1])
            logging.info(f'Injecting sinusoidal signal with power={amplitude} and freq={period}')
            flux = flux + amplitude * np.sin(2 * np.pi * self.time_series[0] / period + random.uniform(0, 2 * np.pi))
        return self.time_series[0], self.time_series[1] + flux - 1, flux

class PseudoPeriodicInjector(FeaturesInjector):
    def __init__(self, time_series: ndarray, amplitude: tuple[float, float], period: tuple[float, float],
                 amplitude_std: tuple[float, float], period_std: tuple[float, float],
                 number_of_signals: int):
        super().__init__(time_series)
        self.power = amplitude
        self.period = period
        self.number_of_signals = number_of_signals
        self.power_std = amplitude_std
        self.period_std = period_std

    def inject(self) -> tuple[ndarray, ndarray, ndarray]:
        number_of_signals = random.uniform(0, 100)
        if number_of_signals < 50:
            number_of_signals = 0
        elif number_of_signals < 75:
            number_of_signals = 1
        elif number_of_signals < 90:
            number_of_signals = 2
        else:
            number_of_signals = int(random.uniform(0, self.number_of_signals))
        flux: ndarray = np.ones(self.time_series.shape[1])
        for index in range(0, number_of_signals):
            period: float = self.random_number_generator.uniform(self.period[0], self.period[1])
            amplitude: float = self.random_number_generator.uniform(self.power[0], self.power[1])
            period_std: float = self.random_number_generator.uniform(self.period_std[0], self.period_std[1])
            amplitude_std: float = self.random_number_generator.uniform(0, amplitude / 10)
            logging.info(f'Injecting pseudo-periodic signal with power={amplitude} and freq={period}')
            pseudo_periodic_signal = timesynth.signals.PseudoPeriodic(amplitude=amplitude, frequency=period, ampSD=amplitude_std, freqSD=0)
            timeseries_pp: ndarray = np.zeros(self.time_series.shape[1])
            for time_index, value in enumerate(self.time_series[0]):
                rn_value = pseudo_periodic_signal.sample_next(value, None, None)
                rn_value = rn_value[0] if isinstance(rn_value, (list, np.ndarray)) else rn_value
                timeseries_pp[time_index] = rn_value
            flux = flux + timeseries_pp
        return self.time_series[0], self.time_series[1] + flux - 1, flux


class TransitsInjector(FeaturesInjector):

    def __init__(self, time_series: ndarray, transit_period_avg_std: tuple[float, float], transit_depth_boundaries: tuple[float, float]):
        super().__init__(time_series)
        self.transit_period_avg_std = transit_period_avg_std
        self.transit_depth_boundaries = transit_depth_boundaries

    def inject(self) -> tuple[ndarray, ndarray, ndarray]:
        number_of_planets = random.uniform(0, 100)
        if number_of_planets < 50:
            number_of_planets = 1
        elif number_of_planets < 80:
            number_of_planets = 2
        else:
            number_of_planets = 3
        star_radius = 1
        star_mass = 1
        flux = np.ones(self.time_series.shape[1])
        for index in range(0, number_of_planets):
            period = 0
            while period < 0.5:
                period = random.gauss(self.transit_period_avg_std[0], self.transit_period_avg_std[1])
            t0 = self.time_series[0][0] + period / random.uniform(0, 1)
            if t0 > self.time_series[0][-1]:
                t0 = np.random.uniform(self.time_series[0][0], self.time_series[0][-1])
            depth = random.uniform(self.transit_depth_boundaries[0], self.transit_depth_boundaries[1])
            planet_radius = np.sqrt(depth * (star_radius ** 2))
            logging.info(f'Injecting planet with P={period}, D={depth}')
            P1 = period * u.day
            a = np.cbrt((ac.G * star_mass * u.M_sun * P1 ** 2) / (4 * np.pi ** 2)).to(u.au)
            ld = [0.5, 0.5]
            planet_model = ellc.lc(
                t_obs=self.time_series[0],
                radius_1=(star_radius * u.R_sun).to(u.au) / a,  # star radius convert from AU to in units of a
                radius_2=(planet_radius * u.R_sun).to(u.au) / a,
                # convert from Rearth (equatorial) into AU and then into units of a
                sbratio=0,
                incl=90,
                light_3=0,
                t_zero=t0,
                period=period,
                a=None,
                q=planet_radius / star_radius * 0.1,
                f_c=None, f_s=None,
                ldc_1=ld, ldc_2=None,
                gdc_1=None, gdc_2=None,
                didt=None,
                domdt=None,
                rotfac_1=1, rotfac_2=1,
                hf_1=1.5, hf_2=1.5,
                bfac_1=None, bfac_2=None,
                heat_1=None, heat_2=None,
                lambda_1=None, lambda_2=None,
                vsini_1=None, vsini_2=None,
                t_exp=None, n_int=None,
                grid_1='default', grid_2='default',
                ld_1='quad', ld_2=None,
                shape_1='sphere', shape_2='sphere',
                spots_1=None, spots_2=None,
                exact_grav=False, verbose=1)
            flux = flux + planet_model - 1
        return self.time_series[0], self.time_series[1] + flux - 1, flux


class FlaresInjector(FeaturesInjector):
    ingress_param = 0.075
    egress_param = 0.15

    def __init__(self, time_series: ndarray, amplitude_avg_std: tuple[float, float], length_avg_std: tuple[float, float]):
        super().__init__(time_series)
        self.amplitude_avg_std = amplitude_avg_std
        self.length_avg_std = length_avg_std

    def inject(self) -> tuple[ndarray, ndarray, ndarray]:
        number_of_flares = random.uniform(0, 100)
        if number_of_flares < 50:
            number_of_flares = 1
        elif number_of_flares < 80:
            number_of_flares = 2
        else:
            number_of_flares = 3
        flux = np.ones(self.time_series.shape[1])
        for index in range(0, number_of_flares):
            t0 = random.choice(self.time_series[0])
            amplitude = random.gauss(self.amplitude_avg_std[0], self.amplitude_avg_std[1])
            length = random.gauss(self.length_avg_std[0], self.length_avg_std[1])
            indexes = np.argwhere((self.time_series[0] > t0 - length // 2) & (self.time_series[0] < t0 + length // 2))
            for i in indexes:
                time_value = self.time_series[0][i]
                if flux[i] < 1 and time_value <= t0:
                    flux[i] = flux[i] + amplitude * np.exp(-((t0 - time_value) ** 2 / (2 * (self.ingress_param ** 2))))
                elif time_value > t0 and flux[i] < 1:
                    flux[i] = flux[i] + amplitude * np.exp((t0 - time_value) / self.egress_param)
        return self.time_series[0], self.time_series[1] + flux - 1, flux

class GapsInjector(FeaturesInjector):
    def __init__(self, time_series: ndarray,
                 gaps_size_boundaries: tuple[float, float], curve_cadence):
        super().__init__(time_series)
        self.gaps_size_boundaries = gaps_size_boundaries
        self.curve_cadence = curve_cadence

    def inject(self) -> tuple[ndarray, ndarray, ndarray]:
        number_of_gaps = random.uniform(0, 100)
        if number_of_gaps < 50:
            number_of_gaps = 1
        elif number_of_gaps < 80:
            number_of_gaps = 2
        else:
            number_of_gaps = 3
        mask_indexes = []
        for _ in np.arange(0, number_of_gaps):
            gap_size = np.random.uniform(self.gaps_size_boundaries[0], self.gaps_size_boundaries[1])
            gap_size_cadences = int(gap_size / self.curve_cadence)
            gap_start_position = int(np.random.uniform(0, self.time_series.shape[1]))
            new_gap_indexes = np.arange(gap_start_position, min(gap_start_position + gap_size_cadences, self.time_series.shape[1]))
            mask_indexes = list(set(mask_indexes).union(new_gap_indexes))
        mask = np.ones(self.time_series.shape[1], dtype=bool)
        mask[mask_indexes] = False
        flux = self.time_series[1]
        flux_model = self.time_series[2]
        flux[~mask] = 0
        flux_model[~mask] = 0
        return self.time_series[0], flux, flux_model


class SantoTestDataGenerator:
    random_number_generator = default_rng()

    def __init__(self, output_dir: str, minimum_curve_size: int,
                 curves_time_boundaries: Tuple[float, float],
                 transit_depth_boundaries: Tuple[float, float],
                 transit_period_avg_std: Tuple[float, float],
                 white_noise_power_boundaries: Tuple[float, float],
                 red_noise_frequency_boundaries: Tuple[float, float],
                 red_noise_power_boundaries: Tuple[float, float],
                 sinusoid_frequency_boundaries: Tuple[float, float],
                 sinusoid_power_boundaries: Tuple[float, float],
                 pseudo_periodic_frequency_boundaries: Tuple[float, float],
                 pseudo_periodic_power_boundaries: Tuple[float, float],
                 pseudo_periodic_frequency_std_boundaries: Tuple[float, float],
                 pseudo_periodic_power_std_boundaries: Tuple[float, float],
                 flare_amplitude_avg_std: Tuple[float, float],
                 flare_length_avg_std: Tuple[float, float],
                 gaps_size_boundaries: Tuple[float, float],
                 max_number_of_pseudo_periodic: int = 3,
                 max_number_of_sinusoid: int = 3,
                 max_number_of_red_noise: int = 5,
                 curves_count=10000, curves_cadences: list=[60, 120, 300, 600, 1800],
                 time_encoding_periods=[5], only_plot=False):
        for curve_index in range(0, curves_count):
            curve_cadences = 1
            while curve_cadences < minimum_curve_size:
                curve_timespan: float = random.uniform(curves_time_boundaries[0], curves_time_boundaries[1])
                selected_cadence = random.choice(curves_cadences) / 3600 / 24
                curve_cadences = int(curve_timespan // selected_cadence)
            flux: ndarray = np.ones((5, curve_cadences))
            flux[0] = np.linspace(0, curve_timespan, curve_cadences)
            white_noise_std: float = 0.0
            while white_noise_std < 0.00001:
                white_noise_std = random.gauss(white_noise_power_boundaries[0], white_noise_power_boundaries[1])
            logging.info(f'Injecting white noise with std={white_noise_std}')
            flux[1] = flux[1] + np.random.normal(loc=0, scale=white_noise_std, size=curve_cadences)
            flux[0], flux[1], flux[2] = RedNoiseInjector(flux, red_noise_power_boundaries, red_noise_frequency_boundaries, max_number_of_red_noise).inject()
            #flux[0] = SinusoidalInjector(time, flux[0], sinusoid_power_boundaries, sinusoid_frequency_boundaries, max_number_of_sinusoid).inject()
            flux[0], flux[1], flux[2] = PseudoPeriodicInjector(flux, pseudo_periodic_power_boundaries, pseudo_periodic_frequency_boundaries,
                                             pseudo_periodic_power_std_boundaries, pseudo_periodic_frequency_std_boundaries, max_number_of_pseudo_periodic).inject()
            flux[0], flux[1], flux[2] = FlaresInjector(flux, flare_amplitude_avg_std, flare_length_avg_std).inject()
            flux[0], flux[1], flux[2] = TransitsInjector(flux, transit_period_avg_std, transit_depth_boundaries).inject()
            flux[0], flux[1], flux[2] = GapsInjector(flux, gaps_size_boundaries, selected_cadence).inject()
            flux[3] = np.copy(flux[2])
            flux[3, np.argwhere((flux[3] <= 1e-7) | (flux[3] > 1 - 1e-7)).flatten()] = float(0)
            flux[3, np.argwhere(flux[3] > 1e-7).flatten()] = float(1)
            flux[4] = np.zeros(flux[0].shape[0])
            for period in time_encoding_periods:
                flux[4] = flux[4] + (np.sin((2 * np.pi / period) * flux[0]) + 1) / 2
            flux[4] = flux[4] / len(time_encoding_periods)
            if only_plot:
                args_empty = np.argwhere(flux[1] > 0).flatten()
                fig, axs = plt.subplots(1, 1, figsize=(10, 5))
                axs.scatter(flux[0][args_empty], flux[1][args_empty], s=2)
                axs.plot(flux[0][args_empty], flux[2][args_empty], c='orange')
                plt.show()
            else:
                np.savetxt(f'{output_dir}/{curve_index}_model.csv', flux, delimiter=',')




from keras.metrics import Mean


class FixedMean(Mean):
    """Fixed mean as the tf.keras.metrics.Mean tries to compute the Mean with a wrong signature method:
    https://stackoverflow.com/questions/68354367/getting-an-error-when-using-tf-keras-metrics-mean-in-functional-keras-api"""
    def update_state(self, y_true, y_pred, sample_weight=None):
        super().update_state(y_pred, sample_weight=sample_weight)


import logging
import sys
from abc import ABC

import matplotlib.pyplot as plt
import random
from typing import Tuple
import ellc
import numpy as np
from numpy import ndarray
from numpy.random import default_rng
import astropy.units as u
import astropy.constants as ac
import timesynth


class FeaturesInjector(ABC):
    random_number_generator = default_rng()

    def __init__(self, time: ndarray, original_flux: ndarray):
        super().__init__()
        self.time = time
        self.original_flux = original_flux

    def inject(self) -> tuple[ndarray, ndarray]:
        pass

class RedNoiseInjector(FeaturesInjector):
    def __init__(self, time: ndarray, original_flux: ndarray, power: tuple[float, float], period: tuple[float, float], number_of_signals: int):
        super().__init__(time, original_flux)
        self.power = power
        self.period = period
        self.number_of_signals = number_of_signals

    def inject(self) -> tuple[ndarray, ndarray]:
        number_of_red_noise = random.uniform(0, 100)
        if number_of_red_noise < 50:
            number_of_red_noise = 0
        elif number_of_red_noise < 75:
            number_of_red_noise = 1
        elif number_of_red_noise < 90:
            number_of_red_noise = 2
        else:
            number_of_red_noise = int(random.uniform(0, self.number_of_signals))
        flux: ndarray = np.ones(len(self.time))
        for index in range(0, number_of_red_noise):
            red_noise_freq: float = self.random_number_generator.uniform(self.period[0], self.period[1])
            red_noise_power: float = self.random_number_generator.uniform(self.power[0], self.power[1])
            logging.info(f'Injecting red noise with power={red_noise_power} and freq={red_noise_freq}')
            red_noise = timesynth.noise.RedNoise(std=red_noise_power, tau=red_noise_freq)
            timeseries_rn: ndarray = np.zeros(len(self.original_flux))
            for time_index, value in enumerate(self.time):
                rn_value = red_noise.sample_next(value, None, None)
                rn_value = rn_value[0] if isinstance(rn_value, (list, np.ndarray)) else rn_value
                timeseries_rn[time_index] = rn_value
            flux = flux + timeseries_rn
        return self.original_flux + flux - 1, flux


class SinusoidalInjector(FeaturesInjector):
    def __init__(self, time: ndarray, original_flux: ndarray, amplitude: tuple[float, float], period: tuple[float, float], number_of_signals: int):
        super().__init__(time, original_flux)
        self.power = amplitude
        self.period = period
        self.number_of_signals = number_of_signals

    def inject(self) -> tuple[ndarray, ndarray]:
        number_of_signals = random.uniform(0, 100)
        if number_of_signals < 50:
            number_of_signals = 0
        elif number_of_signals < 75:
            number_of_signals = 1
        elif number_of_signals < 90:
            number_of_signals = 2
        else:
            number_of_signals = int(random.uniform(0, self.number_of_signals))
        flux: ndarray = np.ones(len(self.time))
        for index in range(0, number_of_signals):
            period: float = self.random_number_generator.uniform(self.period[0], self.period[1])
            amplitude: float = self.random_number_generator.uniform(self.power[0], self.power[1])
            logging.info(f'Injecting sinusoidal signal with power={amplitude} and freq={period}')
            flux = flux + amplitude * np.sin(2 * np.pi * self.time / period + random.uniform(0, 2 * np.pi))
        return self.original_flux + flux - 1, flux

class PseudoPeriodicInjector(FeaturesInjector):
    def __init__(self, time: ndarray, original_flux: ndarray, amplitude: tuple[float, float], period: tuple[float, float], amplitude_std: tuple[float, float], period_std: tuple[float, float],
                 number_of_signals: int):
        super().__init__(time, original_flux)
        self.power = amplitude
        self.period = period
        self.number_of_signals = number_of_signals
        self.power_std = amplitude_std
        self.period_std = period_std

    def inject(self) -> tuple[ndarray, ndarray]:
        number_of_signals = random.uniform(0, 100)
        if number_of_signals < 50:
            number_of_signals = 0
        elif number_of_signals < 75:
            number_of_signals = 1
        elif number_of_signals < 90:
            number_of_signals = 2
        else:
            number_of_signals = int(random.uniform(0, self.number_of_signals))
        flux: ndarray = np.ones(len(self.time))
        for index in range(0, number_of_signals):
            period: float = self.random_number_generator.uniform(self.period[0], self.period[1])
            amplitude: float = self.random_number_generator.uniform(self.power[0], self.power[1])
            period_std: float = self.random_number_generator.uniform(self.period_std[0], self.period_std[1])
            amplitude_std: float = self.random_number_generator.uniform(0, amplitude / 10)
            logging.info(f'Injecting pseudo-periodic signal with power={amplitude} and freq={period}')
            pseudo_periodic_signal = timesynth.signals.PseudoPeriodic(amplitude=amplitude, frequency=period, ampSD=amplitude_std, freqSD=0)
            timeseries_pp: ndarray = np.zeros(len(self.time))
            for time_index, value in enumerate(self.time):
                rn_value = pseudo_periodic_signal.sample_next(value, None, None)
                rn_value = rn_value[0] if isinstance(rn_value, (list, np.ndarray)) else rn_value
                timeseries_pp[time_index] = rn_value
            flux = flux + timeseries_pp
        return self.original_flux + flux - 1, flux


class TransitsInjector(FeaturesInjector):

    def __init__(self, time: ndarray, original_flux: ndarray, transit_period_avg_std: tuple[float, float], transit_depth_boundaries: tuple[float, float]):
        super().__init__(time, original_flux)
        self.transit_period_avg_std = transit_period_avg_std
        self.transit_depth_boundaries = transit_depth_boundaries

    def inject(self) -> tuple[ndarray, ndarray]:
        number_of_planets = random.uniform(0, 100)
        if number_of_planets < 50:
            number_of_planets = 1
        elif number_of_planets < 80:
            number_of_planets = 2
        else:
            number_of_planets = 3
        star_radius = 1
        star_mass = 1
        flux = np.ones(len(self.time))
        for index in range(0, number_of_planets):
            period = 0
            while period < 0.5:
                period = random.gauss(self.transit_period_avg_std[0], self.transit_period_avg_std[1])
            t0 = self.time[0] + period / random.uniform(0, 1)
            if t0 > self.time[-1]:
                t0 = np.random.uniform(self.time[0], self.time[-1])
            depth = random.uniform(self.transit_depth_boundaries[0], self.transit_depth_boundaries[1])
            planet_radius = np.sqrt(depth * (star_radius ** 2))
            logging.info(f'Injecting planet with P={period}, D={depth}')
            P1 = period * u.day
            a = np.cbrt((ac.G * star_mass * u.M_sun * P1 ** 2) / (4 * np.pi ** 2)).to(u.au)
            ld = [0.5, 0.5]
            planet_model = ellc.lc(
                t_obs=self.time,
                radius_1=(star_radius * u.R_sun).to(u.au) / a,  # star radius convert from AU to in units of a
                radius_2=(planet_radius * u.R_sun).to(u.au) / a,
                # convert from Rearth (equatorial) into AU and then into units of a
                sbratio=0,
                incl=90,
                light_3=0,
                t_zero=t0,
                period=period,
                a=None,
                q=planet_radius / star_radius * 0.1,
                f_c=None, f_s=None,
                ldc_1=ld, ldc_2=None,
                gdc_1=None, gdc_2=None,
                didt=None,
                domdt=None,
                rotfac_1=1, rotfac_2=1,
                hf_1=1.5, hf_2=1.5,
                bfac_1=None, bfac_2=None,
                heat_1=None, heat_2=None,
                lambda_1=None, lambda_2=None,
                vsini_1=None, vsini_2=None,
                t_exp=None, n_int=None,
                grid_1='default', grid_2='default',
                ld_1='quad', ld_2=None,
                shape_1='sphere', shape_2='sphere',
                spots_1=None, spots_2=None,
                exact_grav=False, verbose=1)
            flux = flux + planet_model - 1
        return self.original_flux + flux - 1, flux


class FlaresInjector(FeaturesInjector):
    ingress_param = 0.075
    egress_param = 0.15

    def __init__(self, time: ndarray, original_flux: ndarray, amplitude_avg_std: tuple[float, float], length_avg_std: tuple[float, float]):
        super().__init__(time, original_flux)
        self.amplitude_avg_std = amplitude_avg_std
        self.length_avg_std = length_avg_std

    def inject(self) -> tuple[ndarray, ndarray]:
        number_of_flares = random.uniform(0, 100)
        if number_of_flares < 50:
            number_of_flares = 1
        elif number_of_flares < 80:
            number_of_flares = 2
        else:
            number_of_flares = 3
        flux = np.ones(len(self.time))
        for index in range(0, number_of_flares):
            t0 = random.choice(self.time)
            amplitude = random.gauss(self.amplitude_avg_std[0], self.amplitude_avg_std[1])
            length = random.gauss(self.length_avg_std[0], self.length_avg_std[1])
            indexes = np.argwhere((self.time > t0 - length // 2) & (self.time < t0 + length // 2))
            for i in indexes:
                time_value = self.time[i]
                if flux[i] < 1 and time_value <= t0:
                    flux[i] = flux[i] + amplitude * np.exp(-((t0 - time_value) ** 2 / (2 * (self.ingress_param ** 2))))
                elif time_value > t0 and flux[i] < 1:
                    flux[i] = flux[i] + amplitude * np.exp((t0 - time_value) / self.egress_param)
        return self.original_flux + flux - 1, flux


class SantoTestDataGenerator:
    random_number_generator = default_rng()

    def __init__(self, output_dir: str, curves_len_boundaries: Tuple[int, int],
                 transit_depth_boundaries: Tuple[float, float],
                 transit_period_avg_std: Tuple[float, float],
                 white_noise_power_boundaries: Tuple[float, float],
                 red_noise_frequency_boundaries: Tuple[float, float],
                 red_noise_power_boundaries: Tuple[float, float],
                 sinusoid_frequency_boundaries: Tuple[float, float],
                 sinusoid_power_boundaries: Tuple[float, float],
                 pseudo_periodic_frequency_boundaries: Tuple[float, float],
                 pseudo_periodic_power_boundaries: Tuple[float, float],
                 pseudo_periodic_frequency_std_boundaries: Tuple[float, float],
                 pseudo_periodic_power_std_boundaries: Tuple[float, float],
                 flare_amplitude_avg_std: Tuple[float, float],
                 flare_length_avg_std: Tuple[float, float],
                 max_number_of_pseudo_periodic: int = 3,
                 max_number_of_sinusoid: int = 3,
                 max_number_of_red_noise: int = 5,
                 curves_count=10000, curves_cadence=120, only_plot=False):
        for curve_index in range(0, curves_count):
            curve_len: int = int(random.uniform(curves_len_boundaries[0], curves_len_boundaries[1]))
            flux: ndarray = np.ones((2, curve_len))
            curve_len_days: float = curve_len * curves_cadence / 3600 / 24
            time: ndarray = np.linspace(0, curve_len_days, curve_len)
            white_noise_std: float = random.uniform(white_noise_power_boundaries[0], white_noise_power_boundaries[1])
            flux[0] = flux[0] + np.random.normal(loc=0, scale=white_noise_std, size=curve_len)
            flux[0], flux[1] = RedNoiseInjector(time, flux[0], red_noise_power_boundaries, red_noise_frequency_boundaries, max_number_of_red_noise).inject()
            #flux[0] = SinusoidalInjector(time, flux[0], sinusoid_power_boundaries, sinusoid_frequency_boundaries, max_number_of_sinusoid).inject()
            flux[0], flux[1] = PseudoPeriodicInjector(time, flux[0], pseudo_periodic_power_boundaries, pseudo_periodic_frequency_boundaries,
                                             pseudo_periodic_power_std_boundaries, pseudo_periodic_frequency_std_boundaries, max_number_of_pseudo_periodic).inject()
            flux[0], flux[1] = FlaresInjector(time, flux[0], flare_amplitude_avg_std, flare_length_avg_std).inject()
            flux[0], flux[1] = TransitsInjector(time, flux[0], transit_period_avg_std, transit_depth_boundaries).inject()
            #plt.scatter(time, flux[0])
            #plt.show()
            np.savetxt(f'{output_dir}/{curve_index}_model.csv', flux, delimiter=',')




import keras
import tensorflow as tf
import numpy as np
from keras import layers
from tensorflow.python.ops.gen_nn_ops import LeakyRelu


class TransformerEncoder(layers.Layer):
    def __init__(self, key_dim, num_heads, dropout_rate=0.1):
        super(TransformerEncoder, self).__init__()
        self.key_dim = key_dim
        self.num_heads = num_heads
        self.dropout_rate = dropout_rate

    def build(self, input_shape):
        self.input_dim = input_shape[-1]
        self.att = layers.MultiHeadAttention(num_heads=self.num_heads, key_dim=self.key_dim)
        self.ffn = keras.Sequential(
            [layers.Dense(self.input_dim, activation=LeakyRelu(0.01)), layers.Dense(self.key_dim), ]
        )
        self.att_norm = layers.LayerNormalization(epsilon=1e-6)
        self.ffn_norm = layers.LayerNormalization(epsilon=1e-6)
        self.att_dropout = layers.Dropout(self.dropout_rate)
        self.ffn_dropout = layers.Dropout(self.dropout_rate)

    def call(self, inputs, attention_mask, training):
        attn_output = self.att(inputs, inputs, inputs, mask=attention_mask)
        attn_output = self.att_dropout(attn_output, training=training)
        out1 = self.att_norm(inputs + attn_output)
        ffn_output = self.ffn(out1)
        ffn_output = self.ffn_dropout(ffn_output, training=training)
        return self.ffn_norm(out1 + ffn_output)


class TransformerDecoder(layers.Layer):
    def __init__(self, key_dim, num_heads, dropout_rate=0.1):
        super(TransformerDecoder, self).__init__()
        self.key_dim = key_dim
        self.num_heads = num_heads
        self.dropout_rate = dropout_rate

    def build(self, input_shape):
        self.att1 = layers.MultiHeadAttention(num_heads=self.num_heads, key_dim=self.key_dim)
        self.att1_dropout = layers.Dropout(self.dropout_rate)
        self.att1_norm = layers.LayerNormalization(epsilon=1e-6)
        self.att2 = layers.MultiHeadAttention(num_heads=self.num_heads, key_dim=self.key_dim)
        self.att2_dropout = layers.Dropout(self.dropout_rate)
        self.att2_norm = layers.LayerNormalization(epsilon=1e-6)
        self.ffn = keras.Sequential(
            [layers.Dense(self.input_shape, activation=LeakyRelu(0.01)), layers.Dense(self.key_dim), ]
        )
        self.ffn_dropout = layers.Dropout(self.dropout_rate)
        self.ffn_norm = layers.LayerNormalization(epsilon=1e-6)

    def call(self, inputs, encoder_inputs, attention_mask, decoder_mask, training):
        attn_output = self.att1(inputs, inputs, inputs, attention_mask=attention_mask)
        attn_output = self.att1_dropout(attn_output, training=training)
        norm1_output = self.att1_norm(inputs + attn_output)
        attn_output = self.att2(norm1_output, encoder_inputs, encoder_inputs, decoder_mask)
        attn_output = self.att2_dropout(attn_output, training=training)
        norm2_output = self.att2_norm(norm1_output + attn_output)
        ffn_output = self.ffn(norm2_output)
        ffn_output = self.ffn_dropout(ffn_output, training=training)
        return self.ffn_norm(norm2_output + ffn_output)


class Transformer(layers.Layer):
    """
    Implementation of Vaswani et. al (2017) Transformer design with multihead attention
    """
    def __init__(self, key_dim, num_heads, ffn_units, output_dim, stack_depth=2, dropout_rate=0.1):
        super(Transformer, self).__init__()
        self.key_dim = key_dim
        self.ffn_units = ffn_units
        self.output_dim = output_dim
        self.num_heads = num_heads
        self.dropout_rate = dropout_rate
        self.stack_depth = stack_depth

    def build(self, input_shape):
        self.input_dim = input_shape
        self.encoders = np.array([])
        for i in np.arange(0, self.stack_depth):
            self.encoders.append(TransformerEncoder(self.key_dim, self.num_heads, self.dropout_rate))
        self.decoders = np.array([])
        for i in np.arange(0, self.stack_depth):
            self.decoders.append(TransformerDecoder(self.key_dim, self.num_heads, self.dropout_rate))
        self.average_pooling = layers.GlobalAveragePooling1D()
        self.average_pooling_dropout = layers.Dropout(self.dropout_rate)
        self.pooling_dense = layers.Dense(self.ffn_units, activation=LeakyRelu(0.01))
        self.pooling_dense_dropout = layers.Dropout(self.dropout_rate)
        # The output is linear, in case this is a classification problem the softmax is applied in the loss function
        # TODO I don't see where softmax is applied in the loss function
        self.output_dense = layers.Dense(self.output_dim, activation="linear")

    def call(self, input, predicted_output, attention_mask, training):
        encoder_mask = Transformer.__create_padding_mask(input)
        for encoder in self.encoders:
            encoder_output = encoder(input, encoder_mask, training)
        # As the values to be masked are those == 1, we take the maximum to mask any values included in padding mask or
        # look ahead mask
        look_ahead_and_padding_mask = tf.maximum(
            Transformer.__create_padding_mask(predicted_output),
            Transformer.__create_look_ahead_mask(predicted_output)
        )
        decoder_mask = Transformer.__create_padding_mask(input)
        for decoder in self.decoders:
            decoder_output = decoder(predicted_output, encoder_output, look_ahead_and_padding_mask,
                                     decoder_mask, training)
        linear_proj_output = self.average_pooling(decoder_output)
        linear_proj_output = self.pooling_dense(linear_proj_output)
        linear_proj_output = self.pooling_dense_dropout(linear_proj_output, training=training)
        linear_proj_output = self.output_dense(linear_proj_output, training=training)
        return linear_proj_output

    @staticmethod
    def __create_padding_mask(input_sequence):
        """
        Creates the mask for zero-padded input sequences with dimension (batch_size, seq_length)
        :param input_sequence: 
        :return: the mask redimensioned to ?????
        """
        # TODO clarify the return step where new dimensions are created somehow?
        mask = tf.cast(tf.math.equal(input_sequence, 0), tf.float32)
        return mask[:, tf.newaxis, tf.newaxis, :]

    @staticmethod
    def __create_look_ahead_mask(seq):
        """Creates a mask taking all the elements from the diagonal and above of a tensor of (seq_length, seq_length)
        dimension. This ensures that we are masking the future tokens only:
        1 1 1
        0 1 1
        0 0 1
        """
        # Create the mask for the causal attention
        seq_len = tf.shape(seq)[1]
        look_ahead_mask = 1 - tf.linalg.band_part(tf.ones((seq_len, seq_len)), -1, 0)
        return look_ahead_mask


import logging

import numpy as np
from keras_core import Layer


class ModelWeightsLogger:
    def log_model_weights(self, model):
        # this function runs at the end of each epoch
        #logging.info('Generating layer weights info')
        # loop over each layer and get weights and biases
        min_w = 0
        max_w = 0
        min_layer = ""
        max_layer = ""
        for layer_i in range(len(model.layers)):
            weights_and_bias = model.layers[layer_i].get_weights()
            if len(weights_and_bias) > 0:
                w = weights_and_bias[0]
                b = weights_and_bias[1]
                # logging.debug('Layer no %s of type %s has weights of shape %s and biases of shape %s' % (
                #     layer_i, model.layers[layer_i], np.shape(w), np.shape(b)))
                nan_weight_indexes = np.argwhere(np.isnan(w)).flatten()
                nan_bias_indexes = np.argwhere(np.isnan(w)).flatten()
                nan_weight_indexes_count = len(nan_weight_indexes)
                nan_bias_indexes_count = len(nan_bias_indexes)
                min_layer = min_layer if np.min(w) > min_w else model.layers[layer_i]
                max_layer = max_layer if np.max(w) < max_w else model.layers[layer_i]
                min_w = min_w if np.min(w) > min_w else np.min(w)
                max_w = max_w if np.max(w) < max_w else np.max(w)
                # logging.debug('Layer max weight is ' + str(np.nanmax(w)) + ' and min weight is ' + str(np.nanmin(w)))
                # logging.debug('Layer max bias is ' + str(np.nanmax(b)) + ' and min bias is ' + str(np.nanmin(b)))
                if nan_weight_indexes_count > 0 or nan_bias_indexes_count > 0:
                    logging.warning('Layer ' + str(layer_i) + ' with name ' + model.layers[layer_i].name + ' has ' + str(nan_weight_indexes_count) +
                          ' nan weights and ' + str(nan_bias_indexes_count) + ' nan biases')
        if isinstance(min_layer, Layer):
            logging.info("Min weights: %s from layer %s", min_w, min_layer.name)
        if isinstance(max_layer, Layer):
            logging.info("Max weights: %s from layer %s", max_w, max_layer.name)



import collections
import csv
import logging
import numpy as np
import tensorflow as tf
import pandas

from exoml.ml.callback.training_data_aware import ValidationDataAwareCallback
from exoml.ml.metrics.auc import confusion_matrix_values, area_under_curve


class ConfusionMatrixStatisticsCallback(ValidationDataAwareCallback):
    def __init__(self, filenames, validation_data=None):
        super().__init__(validation_data=validation_data)
        self.filenames = filenames
        self.stats_csv_files = []
        self.writers = []
        self.roc_dfs = []
        self.keys = None
        self.roc_df = pandas.DataFrame(columns=['epoch', 'label', 'precision', 'recall', 'thresholds'])
        mode = 'w'
        for filename in self.filenames:
            self.stats_csv_files = self.stats_csv_files + [tf.io.gfile.GFile(self.model_dir + '/' + filename +
                                                                             '_stats.csv', mode)]

    def _close_resources(self):
        for index, filename in enumerate(self.filenames):
            self.stats_csv_files[index].close()
            self.writers[index] = None
        self.writers = None
        self.roc_dfs = None

    def on_train_end(self, logs=None):
        self._close_resources()

    def on_epoch_end(self, epoch, logs={}):
        # x_train, y_train = self.train[0], self.train[1]
        # y_pred = self.model.predict(x_train)
        # confusion_matrix, accuracy, error_rate, precision, recall, f1_score = self.confusion_matrix(y_train, y_pred)
        # logs['F1_score_train'] = np.round(f1_score, 5)
        # if self.validation:
        logging.info("Computing AUC metrics")
        batches = len(self.validation_data)
        y_valid = None
        for batch in range(batches):
            x_val, y_val = self.validation_data[batch]
            if y_valid is None:
                y_valid = y_val
            else:
                y_valid = np.concatenate((y_valid, y_val))
        y_val_pred = self.model.predict(self.validation_data)
        logs = logs or {}

        def handle_value(k):
            is_zero_dim_ndarray = isinstance(k, np.ndarray) and k.ndim == 0
            if isinstance(k, str):
                return k
            elif isinstance(k, collections.abc.Iterable) and not is_zero_dim_ndarray:
                return '"[%s]"' % (', '.join(map(str, k)))
            else:
                return k

        if self.keys is None:
            self.keys = sorted(logs.keys())
        if self.model.stop_training:
            # We set NA so that csv parsers do not fail for this last epoch.
            logs = dict((k, logs[k]) if k in logs else (k, 'NA') for k in self.keys)
        if len(self.writers) == 0:
            class CustomDialect(csv.excel):
                delimiter = ','
            fieldnames = ['epoch', 'val_f1_score', 'val_tp', 'val_fp', 'val_fn', 'val_tn', 'val_precision',
                          'val_recall', 'val_roc'] + self.keys
            for csv_file in self.stats_csv_files:
                new_writer = csv.DictWriter(csv_file, fieldnames=fieldnames, dialect=CustomDialect)
                new_writer.writeheader()
                self.writers = self.writers + [new_writer]

        for index, filename in enumerate(self.filenames):
            confusion_matrix, accuracy, error_rate, precision, recall, f1_score = \
                confusion_matrix_values(y_valid, y_val_pred, label_index=index)
            recalls, precisions, thresholds, roc = area_under_curve(y_valid, y_val_pred, label_index=index)
            row_dict = collections.OrderedDict({'epoch': epoch, 'val_f1_score': f1_score,
                                                'val_tp': confusion_matrix[0][0],
                                                'val_fp': confusion_matrix[0][1],
                                                'val_fn': confusion_matrix[1][0],
                                                'val_tn': confusion_matrix[1][1],
                                                'val_precision': precision,
                                                'val_recall': recall,
                                                'val_roc': roc
                                                })
            self.writers[index].writerow(row_dict)
            self.stats_csv_files[index].flush()
            for precision_index, precision in enumerate(precisions):
                self.roc_df = self.roc_df.append({'epoch': epoch, 'label': index, 'recall': recalls[precision_index],
                                                  'precision': precisions[precision_index],
                                                  'thresholds': thresholds[precision_index]}, ignore_index=True)
        self.roc_df.to_csv(self.model_dir + '/precision_recall.csv', index=False)
        logging.info("Computed AUC metrics")


import abc
import importlib
import warnings

from keras import backend
from keras_core.src.utils import tf_utils
# from keras.src import backend
# from keras.src.utils import tf_utils

from keras.callbacks import Callback
import tensorflow as tf
from tensorflow_addons.optimizers import MultiOptimizer
from typeguard import typechecked
from typing import Union, List


if (
    hasattr(tf.keras.optimizers, "experimental")
    and tf.keras.optimizers.Optimizer.__module__
    == tf.keras.optimizers.experimental.Optimizer.__module__
):
    # If the default optimizer points to new Keras optimizer, addon optimizers
    # should use the legacy path.
    KerasLegacyOptimizer = tf.keras.optimizers.legacy.Optimizer
else:
    KerasLegacyOptimizer = tf.keras.optimizers.Optimizer


if importlib.util.find_spec("tensorflow.keras.optimizers.legacy") is not None:
    Optimizer = Union[
        tf.keras.optimizers.Optimizer, tf.keras.optimizers.legacy.Optimizer, str
    ]
else:
    Optimizer = Union[tf.keras.optimizers.Optimizer, str]


class AveragedOptimizerWrapper(KerasLegacyOptimizer, metaclass=abc.ABCMeta):
    @typechecked
    def __init__(
        self,
        optimizer: Optimizer,
        name: str = "AverageOptimizer",
        **kwargs,
    ):
        super().__init__(name, **kwargs)

        if isinstance(optimizer, str):
            if (
                hasattr(tf.keras.optimizers, "legacy")
                and KerasLegacyOptimizer == tf.keras.optimizers.legacy.Optimizer
            ):
                optimizer = tf.keras.optimizers.get(
                    optimizer, use_legacy_optimizer=True
                )
            else:
                optimizer = tf.keras.optimizers.get(optimizer)

        if not isinstance(optimizer, KerasLegacyOptimizer):
            raise TypeError(
                "optimizer is not an object of tf.keras.optimizers.legacy.Optimizer "
            )

        self._optimizer = optimizer
        self._track_trackable(self._optimizer, "awg_optimizer")

    def _create_slots(self, var_list):
        self._optimizer._create_slots(var_list=var_list)
        for var in var_list:
            self.add_slot(var, "average")

    def _create_hypers(self):
        self._optimizer._create_hypers()

    def _prepare_local(self, var_device, var_dtype, apply_state):
        return self._optimizer._prepare_local(var_device, var_dtype, apply_state)

    def apply_gradients(self, grads_and_vars, name=None, **kwargs):
        self._optimizer._iterations = self.iterations
        return super().apply_gradients(grads_and_vars, name, **kwargs)

    @abc.abstractmethod
    def average_op(self, var, average_var, local_apply_state):
        raise NotImplementedError

    def _apply_average_op(self, train_op, var, apply_state):
        apply_state = apply_state or {}
        local_apply_state = apply_state.get((var.device, var.dtype.base_dtype))
        if local_apply_state is None:
            local_apply_state = self._fallback_apply_state(
                var.device, var.dtype.base_dtype
            )
        average_var = self.get_slot(var, "average")
        return self.average_op(var, average_var, local_apply_state)

    def _resource_apply_dense(self, grad, var, apply_state=None):
        if "apply_state" in self._optimizer._dense_apply_args:
            train_op = self._optimizer._resource_apply_dense(
                grad, var, apply_state=apply_state
            )
        else:
            train_op = self._optimizer._resource_apply_dense(grad, var)
        average_op = self._apply_average_op(train_op, var, apply_state)
        return tf.group(train_op, average_op)

    def _resource_apply_sparse(self, grad, var, indices, apply_state=None):
        if "apply_state" in self._optimizer._sparse_apply_args:
            train_op = self._optimizer._resource_apply_sparse(
                grad, var, indices, apply_state=apply_state
            )
        else:
            train_op = self._optimizer._resource_apply_sparse(grad, var, indices)
        average_op = self._apply_average_op(train_op, var, apply_state)
        return tf.group(train_op, average_op)

    def _resource_apply_sparse_duplicate_indices(
        self, grad, var, indices, apply_state=None
    ):
        if "apply_state" in self._optimizer._sparse_apply_args:
            train_op = self._optimizer._resource_apply_sparse_duplicate_indices(
                grad, var, indices, apply_state=apply_state
            )
        else:
            train_op = self._optimizer._resource_apply_sparse_duplicate_indices(
                grad, var, indices
            )
        average_op = self._apply_average_op(train_op, var, apply_state)
        return tf.group(train_op, average_op)

    def assign_average_vars(self, var_list):
        """Assign variables in var_list with their respective averages.

        Args:
            var_list: List of model variables to be assigned to their average.

        Returns:
            assign_op: The op corresponding to the assignment operation of
            variables to their average.

        Example:
        ```python
        model = tf.Sequential([...])
        opt = tfa.optimizers.SWA(
                tf.keras.optimizers.SGD(lr=2.0), 100, 10)
        model.compile(opt, ...)
        model.fit(x, y, ...)

        # Update the weights to their mean before saving
        opt.assign_average_vars(model.variables)

        model.save('model.h5')
        ```
        """
        assign_ops = []
        for var in var_list:
            try:
                assign_ops.append(
                    var.assign(
                        self.get_slot(var, "average"),
                        use_locking=self._use_locking,
                    )
                )
            except Exception as e:
                warnings.warn("Unable to assign average slot to {} : {}".format(var, e))
        return tf.group(assign_ops)

    def get_config(self):
        config = {
            "optimizer": tf.keras.optimizers.serialize(self._optimizer),
        }
        base_config = super().get_config()
        return {**base_config, **config}

    @classmethod
    def from_config(cls, config, custom_objects=None):
        optimizer = tf.keras.optimizers.deserialize(
            config.pop("optimizer"), custom_objects=custom_objects
        )
        return cls(optimizer, **config)

    @property
    def weights(self):
        return self._weights + self._optimizer.weights

    @property
    def lr(self):
        return self._optimizer._get_hyper("learning_rate")

    @lr.setter
    def lr(self, lr):
        self._optimizer._set_hyper("learning_rate", lr)  #

    @property
    def learning_rate(self):
        return self._optimizer._get_hyper("learning_rate")

    @learning_rate.setter
    def learning_rate(self, learning_rate):
        self._optimizer._set_hyper("learning_rate", learning_rate)


# @tf.keras.utils.register_keras_serializable(package="Addons")
# class SWA(AveragedOptimizerWrapper):
#     """This class extends optimizers with Stochastic Weight Averaging (SWA).
#
#     The Stochastic Weight Averaging mechanism was proposed by Pavel Izmailov
#     et. al in the paper [Averaging Weights Leads to Wider Optima and
#     Better Generalization](https://arxiv.org/abs/1803.05407). The optimizer
#     implements averaging of multiple points along the trajectory of SGD. The
#     optimizer expects an inner optimizer which will be used to apply the
#     gradients to the variables and itself computes a running average of the
#     variables every `k` steps (which generally corresponds to the end
#     of a cycle when a cyclic learning rate is employed).
#
#     We also allow the specification of the number of steps averaging
#     should first happen after. Let's say, we want averaging to happen every `k`
#     steps after the first `m` steps. After step `m` we'd take a snapshot of the
#     variables and then average the weights appropriately at step `m + k`,
#     `m + 2k` and so on. The assign_average_vars function can be called at the
#     end of training to obtain the averaged_weights from the optimizer.
#
#     Note: If your model has batch-normalization layers you would need to run
#     the final weights through the data to compute the running mean and
#     variance corresponding to the activations for each layer of the network.
#     From the paper: If the DNN uses batch normalization we run one
#     additional pass over the data, to compute the running mean and standard
#     deviation of the activations for each layer of the network with SWA
#     weights after the training is finished, since these statistics are not
#     collected during training. For most deep learning libraries, such as
#     PyTorch or Tensorflow, one can typically collect these statistics by
#     making a forward pass over the data in training mode
#     ([Averaging Weights Leads to Wider Optima and Better
#     Generalization](https://arxiv.org/abs/1803.05407))
#
#     Example of usage:
#
#     ```python
#     opt = tf.keras.optimizers.SGD(learning_rate)
#     opt = tfa.optimizers.SWA(opt, start_averaging=m, average_period=k)
#     ```
#     """
#
#     @typechecked
#     def __init__(
#         self,
#         optimizer: Optimizer,
#         start_averaging: int = 0,
#         average_period: int = 10,
#         name: str = "SWA",
#         **kwargs,
#     ):
#         r"""Wrap optimizer with the Stochastic Weight Averaging mechanism.
#
#         Args:
#             optimizer: The original optimizer that will be used to compute and
#                 apply the gradients.
#             start_averaging: An integer. Threshold to start averaging using
#                 SWA. Averaging only occurs at `start_averaging` iters, must
#                 be >= 0. If start_averaging = m, the first snapshot will be
#                 taken after the mth application of gradients (where the first
#                 iteration is iteration 0).
#             average_period: An integer. The synchronization period of SWA. The
#                 averaging occurs every average_period steps. Averaging period
#                 needs to be >= 1.
#             name: Optional name for the operations created when applying
#                 gradients. Defaults to 'SWA'.
#             **kwargs: keyword arguments. Allowed to be {`clipnorm`,
#                 `clipvalue`, `lr`, `decay`}. `clipnorm` is clip gradients by
#                 norm; `clipvalue` is clip gradients by value, `decay` is
#                 included for backward compatibility to allow time inverse
#                 decay of learning rate. `lr` is included for backward
#                 compatibility, recommended to use `learning_rate` instead.
#         """
#         super().__init__(optimizer, name, **kwargs)
#
#         if average_period < 1:
#             raise ValueError("average_period must be >= 1")
#         if start_averaging < 0:
#             raise ValueError("start_averaging must be >= 0")
#
#         self._set_hyper("average_period", average_period)
#         self._set_hyper("start_averaging", start_averaging)
#
#     @tf.function
#     def average_op(self, var, average_var, local_apply_state):
#         average_period = self._get_hyper("average_period", tf.dtypes.int64)
#         start_averaging = self._get_hyper("start_averaging", tf.dtypes.int64)
#         # number of times snapshots of weights have been taken (using max to
#         # avoid negative values of num_snapshots).
#         num_snapshots = tf.math.maximum(
#             tf.cast(0, tf.int64),
#             tf.math.floordiv(self.iterations - start_averaging, average_period),
#         )
#
#         # The average update should happen iff two conditions are met:
#         # 1. A min number of iterations (start_averaging) have taken place.
#         # 2. Iteration is one in which snapshot should be taken.
#         checkpoint = start_averaging + num_snapshots * average_period
#         if self.iterations >= start_averaging and self.iterations == checkpoint:
#             num_snapshots = tf.cast(num_snapshots, tf.float32)
#             average_value = (average_var * num_snapshots + var) / (num_snapshots + 1.0)
#             return average_var.assign(average_value, use_locking=self._use_locking)
#
#         return average_var
#
#     def get_config(self):
#         config = {
#             "average_period": self._serialize_hyperparameter("average_period"),
#             "start_averaging": self._serialize_hyperparameter("start_averaging"),
#         }
#         base_config = super().get_config()
#         return {**base_config, **config}


# @tf.keras.utils.register_keras_serializable(package="Addons")
# class MultiOptimizer(KerasLegacyOptimizer):
#     """Multi Optimizer Wrapper for Discriminative Layer Training.
#
#     Creates a wrapper around a set of instantiated optimizer layer pairs.
#     Generally useful for transfer learning of deep networks.
#
#     Each optimizer will optimize only the weights associated with its paired layer.
#     This can be used to implement discriminative layer training by assigning
#     different learning rates to each optimizer layer pair.
#     `(tf.keras.optimizers.legacy.Optimizer, List[tf.keras.layers.Layer])` pairs are also supported.
#     Please note that the layers must be instantiated before instantiating the optimizer.
#
#     Args:
#         optimizers_and_layers: a list of tuples of an optimizer and a layer or model.
#             Each tuple should contain exactly 1 instantiated optimizer and 1 object that
#             subclasses `tf.keras.Model`, `tf.keras.Sequential` or `tf.keras.layers.Layer`.
#             Nested layers and models will be automatically discovered.
#             Alternatively, in place of a single layer, you can pass a list of layers.
#         optimizer_specs: specialized list for serialization.
#             Should be left as None for almost all cases.
#             If you are loading a serialized version of this optimizer,
#             please use `tf.keras.models.load_model` after saving a model compiled with this optimizer.
#
#     Usage:
#
#     >>> model = tf.keras.Sequential([
#     ...     tf.keras.Input(shape=(4,)),
#     ...     tf.keras.layers.Dense(8),
#     ...     tf.keras.layers.Dense(16),
#     ...     tf.keras.layers.Dense(32),
#     ... ])
#     >>> optimizers = [
#     ...     tf.keras.optimizers.Adam(learning_rate=1e-4),
#     ...     tf.keras.optimizers.Adam(learning_rate=1e-2)
#     ... ]
#     >>> optimizers_and_layers = [(optimizers[0], model.layers[0]), (optimizers[1], model.layers[1:])]
#     >>> optimizer = tfa.optimizers.MultiOptimizer(optimizers_and_layers)
#     >>> model.compile(optimizer=optimizer, loss="mse")
#
#     Reference:
#         - [Universal Language Model Fine-tuning for Text Classification](https://arxiv.org/abs/1801.06146)
#         - [Collaborative Layer-wise Discriminative Learning in Deep Neural Networks](https://arxiv.org/abs/1607.05440)
#
#     Note: Currently, `tfa.optimizers.MultiOptimizer` does not support callbacks that modify optimizers.
#         However, you can instantiate optimizer layer pairs with
#         `tf.keras.optimizers.schedules.LearningRateSchedule`
#         instead of a static learning rate.
#
#     This code should function on CPU, GPU, and TPU. Apply with `tf.distribute.Strategy().scope()` context as you
#     would with any other optimizer.
#     """
#
#     @typechecked
#     def __init__(
#         self,
#         optimizers_and_layers: Union[list, None] = None,
#         optimizer_specs: Union[list, None] = None,
#         name: str = "MultiOptimizer",
#         **kwargs,
#     ):
#
#         super(MultiOptimizer, self).__init__(name, **kwargs)
#
#         if optimizer_specs is None and optimizers_and_layers is not None:
#             self.optimizer_specs = [
#                 self.create_optimizer_spec(optimizer, layers_or_model)
#                 for optimizer, layers_or_model in optimizers_and_layers
#             ]
#
#         elif optimizer_specs is not None and optimizers_and_layers is None:
#             self.optimizer_specs = [
#                 self.maybe_initialize_optimizer_spec(spec) for spec in optimizer_specs
#             ]
#
#         else:
#             raise RuntimeError(
#                 "Must specify one of `optimizers_and_layers` or `optimizer_specs`."
#             )
#
#     def apply_gradients(self, grads_and_vars, **kwargs):
#         """Wrapped apply_gradient method.
#
#         Returns an operation to be executed.
#         """
#
#         for spec in self.optimizer_specs:
#             spec["gv"] = []
#
#         for grad, var in tuple(grads_and_vars):
#             for spec in self.optimizer_specs:
#                 for name in spec["weights"]:
#                     if var.name == name:
#                         spec["gv"].append((grad, var))
#
#         update_ops = [
#             spec["optimizer"].apply_gradients(spec["gv"], **kwargs)
#             for spec in self.optimizer_specs
#         ]
#         update_group = tf.group(update_ops)
#
#         any_symbolic = any(
#             isinstance(i, tf.Operation) or tf_utils.is_symbolic_tensor(i)
#             for i in update_ops
#         )
#
#         if not tf.executing_eagerly() or any_symbolic:
#             # If the current context is graph mode or any of the update ops are
#             # symbolic then the step update should be carried out under a graph
#             # context. (eager updates execute immediately)
#             with backend._current_graph(  # pylint: disable=protected-access
#                 update_ops
#             ).as_default():
#                 with tf.control_dependencies([update_group]):
#                     return self.iterations.assign_add(1, read_value=False)
#
#         return self.iterations.assign_add(1)
#
#     def get_config(self):
#         config = super(MultiOptimizer, self).get_config()
#         optimizer_specs_without_gv = []
#         for optimizer_spec in self.optimizer_specs:
#             optimizer_specs_without_gv.append(
#                 {
#                     "optimizer": optimizer_spec["optimizer"],
#                     "weights": optimizer_spec["weights"],
#                 }
#             )
#         config.update({"optimizer_specs": optimizer_specs_without_gv})
#         return config
#
#     @classmethod
#     def create_optimizer_spec(
#         cls,
#         optimizer: KerasLegacyOptimizer,
#         layers_or_model: Union[
#             tf.keras.Model,
#             tf.keras.Sequential,
#             tf.keras.layers.Layer,
#             List[tf.keras.layers.Layer],
#         ],
#     ):
#         """Creates a serializable optimizer spec.
#
#         The name of each variable is used rather than `var.ref()` to enable serialization and deserialization.
#         """
#         if isinstance(layers_or_model, list):
#             weights = [
#                 var.name for sublayer in layers_or_model for var in sublayer.weights
#             ]
#         else:
#             weights = [var.name for var in layers_or_model.weights]
#
#         return {
#             "optimizer": optimizer,
#             "weights": weights,
#         }
#
#     @classmethod
#     def maybe_initialize_optimizer_spec(cls, optimizer_spec):
#         if isinstance(optimizer_spec["optimizer"], dict):
#             optimizer_spec["optimizer"] = tf.keras.optimizers.deserialize(
#                 optimizer_spec["optimizer"]
#             )
#
#         return optimizer_spec
#
#     def __repr__(self):
#         return "Multi Optimizer with %i optimizer layer pairs" % len(
#             self.optimizer_specs
#         )


class WarmUpAndLinDecreaseCallback(Callback):
    def __init__(
            self,
            initial_lr,
            top_lr,
            baseline_lr,
            warmup_epochs,
            baseline_epochs,
            name=None):
        """Applies linear decrease with a first stage of warmup
    """
        super(WarmUpAndLinDecreaseCallback, self).__init__()
        assert initial_lr > 0
        self.initial_lr = initial_lr
        self.top_lr = top_lr
        self.baseline_lr = baseline_lr
        self.warmup_epochs = warmup_epochs
        self.baseline_epochs = baseline_epochs
        self.name = name
        super().__init__()

    def on_epoch_end(self, epoch, logs=None):
        if self.model.optimizer.__class__.__name__ == MultiOptimizer.__name__:
            for index, optimizer_spec in enumerate(self.model.optimizer.optimizer_specs):
                optimizer = optimizer_spec['optimizer']
                warmup_slope = (self.top_lr - self.initial_lr) / self.warmup_epochs
                decreasing_slope = (self.baseline_lr - self.top_lr) / self.baseline_epochs
                new_lr_decreasing = epoch * decreasing_slope + self.top_lr
                new_lr_warmup = epoch * warmup_slope + self.initial_lr
                new_learning_rate = new_lr_warmup if epoch <= self.warmup_epochs else new_lr_decreasing
                new_learning_rate = new_learning_rate if new_learning_rate < self.baseline_lr else new_learning_rate
                new_learning_rate = new_learning_rate * optimizer.progressive_lr_factor
                backend.set_value(optimizer.lr, new_learning_rate)
        else:
            optimizer = self.model.optimizer
            if self.warmup_epochs <= 0:
                warmup_slope = 1
            else:
                warmup_slope = (self.top_lr - self.initial_lr) / self.warmup_epochs
            decreasing_slope = (self.baseline_lr - self.top_lr) / self.baseline_epochs
            new_lr_decreasing = epoch * decreasing_slope + self.top_lr
            new_lr_warmup = epoch * warmup_slope + self.initial_lr
            new_learning_rate = new_lr_warmup if epoch <= self.warmup_epochs else new_lr_decreasing
            new_learning_rate = new_learning_rate if new_learning_rate < self.baseline_lr else new_learning_rate
            backend.set_value(optimizer.lr, new_learning_rate)



import logging
import os
import pathlib
import re
import shutil
import sys
from multiprocessing import Pool

import foldedleastsquares
import sklearn
import wotan
from sshkeyboard import listen_keyboard, stop_listening
import keras
import pandas as pd
import lightkurve as lk
import foldedleastsquares as tls
import matplotlib.pyplot as plt
import astropy.units as u
import requests
from astropy.coordinates import SkyCoord
from astropy.wcs import WCS
from astroquery.mast import Catalogs, Tesscut
from keras import Sequential
from keras.layers import Conv1D, Dropout, MaxPooling1D, Flatten, Dense, Embedding, LSTM, ConvLSTM2D
from lightkurve import TessTargetPixelFile
from matplotlib.gridspec import GridSpec

from sherlockpipe.ois.OisManager import OisManager

from lcbuilder.objectinfo.MissionObjectInfo import MissionObjectInfo
from lcbuilder.objectinfo.preparer.MissionLightcurveBuilder import MissionLightcurveBuilder
from lcbuilder.eleanor import TargetData
from lcbuilder import eleanor
from lcbuilder.star.TicStarCatalog import TicStarCatalog
import numpy as np
# import tsfresh
# from tsfresh.utilities.dataframe_functions import impute

class PrepareTicInput:
    def __init__(self, dir, tic, target_ois, sector):
        self.dir = dir
        self.tic = tic
        self.target_ois = target_ois
        self.sector = sector

class EmptyTargetsPreparer:
    SECTOR_URL = "https://tess.mit.edu/wp-content/uploads/all_targets_S0{:02}_v1.csv"

    def __init__(self, dir, cache_dir):
        self.dir = dir
        if not os.path.exists(dir):
            os.mkdir(dir)
        file_dir = dir + "/ml.log"
        if os.path.exists(file_dir):
            os.remove(file_dir)
        formatter = logging.Formatter('%(message)s')
        logger = logging.getLogger()
        while len(logger.handlers) > 0:
            logger.handlers.pop()
        logger.setLevel(logging.INFO)
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(logging.INFO)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        handler = logging.FileHandler(file_dir)
        handler.setLevel(logging.INFO)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        self.positive_dir = dir + "/tp/"
        self.negative_dir = dir + "/ntp/"
        self.false_positive_dir = dir + "/fp/"
        if not os.path.isdir(dir):
            os.mkdir(dir)
        if not os.path.isdir(self.positive_dir):
            os.mkdir(self.positive_dir)
        if not os.path.isdir(self.negative_dir):
            os.mkdir(self.negative_dir)
        if not os.path.isdir(self.false_positive_dir):
            os.mkdir(self.false_positive_dir)

    def download_neighbours(self, ID: int, sectors: np.ndarray, search_radius: int = 10):
        """
        Queries TIC for sources near the target and obtains a cutout
        of the pixels enclosing the target.
        Args:
            ID (int): TIC ID of the target.
            sectors (numpy array): Sectors in which the target
                                   has been observed.
            search_radius (int): Number of pixels from the target
                                 star to search.
        """
        ID = ID
        sectors = sectors
        search_radius = search_radius
        N_pix = 2 * search_radius + 2
        # query TIC for nearby stars
        pixel_size = 20.25 * u.arcsec
        df = Catalogs.query_object(
            str(ID),
            radius=search_radius * pixel_size,
            catalog="TIC"
        )
        new_df = df[
            "ID", "Tmag", "ra", "dec", "mass", "rad", "Teff", "plx"
        ]
        stars = new_df.to_pandas()

        TESS_images = []
        col0s, row0s = [], []
        pix_coords = []
        # for each sector, get FFI cutout and transform RA/Dec into
        # TESS pixel coordinates
        for j, sector in enumerate(sectors):
            Tmag = stars["Tmag"].values
            ra = stars["ra"].values
            dec = stars["dec"].values
            cutout_coord = SkyCoord(ra[0], dec[0], unit="deg")
            cutout_hdu = Tesscut.get_cutouts(cutout_coord, size=N_pix, sector=sector)[0]
            cutout_table = cutout_hdu[1].data
            hdu = cutout_hdu[2].header
            wcs = WCS(hdu)
            TESS_images.append(np.mean(cutout_table["FLUX"], axis=0))
            col0 = cutout_hdu[1].header["1CRV4P"]
            row0 = cutout_hdu[1].header["2CRV4P"]
            col0s.append(col0)
            row0s.append(row0)

            pix_coord = np.zeros([len(ra), 2])
            for i in range(len(ra)):
                RApix = np.asscalar(
                    wcs.all_world2pix(ra[i], dec[i], 0)[0]
                )
                Decpix = np.asscalar(
                    wcs.all_world2pix(ra[i], dec[i], 0)[1]
                )
                pix_coord[i, 0] = col0 + RApix
                pix_coord[i, 1] = row0 + Decpix
            pix_coords.append(pix_coord)

        # for each star, get the separation and position angle
        # from the targets star
        sep = [0]
        pa = [0]
        c_target = SkyCoord(
            stars["ra"].values[0],
            stars["dec"].values[0],
            unit="deg"
        )
        for i in range(1, len(stars)):
            c_star = SkyCoord(
                stars["ra"].values[i],
                stars["dec"].values[i],
                unit="deg"
            )
            sep.append(
                np.round(
                    c_star.separation(c_target).to(u.arcsec).value,
                    3
                )
            )
            pa.append(
                np.round(
                    c_star.position_angle(c_target).to(u.deg).value,
                    3
                )
            )
        stars["sep (arcsec)"] = sep
        stars["PA (E of N)"] = pa

        stars = stars
        TESS_images = TESS_images
        col0s = col0s
        row0s = row0s
        pix_coords = pix_coords
        return stars

    def store_lc_data(self, lc_data, file):
        # time = lc_data["time"].to_numpy()
        # dif = time[1:] - time[:-1]
        # jumps = np.where(np.abs(dif) > 0.5)[0]
        # jumps = np.append(jumps, len(lc_data))
        # previous_jump_index = 0
        # for jumpIndex in jumps:
        #     token = lc_data["centroids_x"][previous_jump_index:jumpIndex].to_numpy()
        #     lc_data["centroids_x"][previous_jump_index:jumpIndex] = np.tanh(token - np.nanmedian(token))
        #     token = lc_data["centroids_y"][previous_jump_index:jumpIndex].to_numpy()
        #     lc_data["centroids_y"][previous_jump_index:jumpIndex] = np.tanh(token - np.nanmedian(token))
        #     token = lc_data["motion_x"][previous_jump_index:jumpIndex].to_numpy()
        #     lc_data["motion_x"][previous_jump_index:jumpIndex] = np.tanh(token - np.nanmedian(token))
        #     token = lc_data["motion_y"][previous_jump_index:jumpIndex].to_numpy()
        #     lc_data["motion_y"][previous_jump_index:jumpIndex] = np.tanh(token - np.nanmedian(token))
        #     token = lc_data["background_flux"][previous_jump_index:jumpIndex].to_numpy()
        #     lc_data["background_flux"][previous_jump_index:jumpIndex] = np.tanh(token - np.nanmedian(token))
        #     previous_jump_index = jumpIndex
        lc_data.to_csv(file)
        return lc_data

    def prepare_tic(self, prepare_tic_input: PrepareTicInput):
        tic_id = str(prepare_tic_input.tic)
        target_dir = prepare_tic_input.dir
        if not os.path.isdir(target_dir):
            os.mkdir(target_dir)
        try:
            mission_lightcurve_builder = MissionLightcurveBuilder()
        except Exception as e:
            logging.exception("Failed target %s for sector %s", tic_id, prepare_tic_input.sector)
        try:
            logging.info("Trying to get short cadence info for " + str(prepare_tic_input.tic))
            lcbuild = \
                mission_lightcurve_builder.build(MissionObjectInfo([prepare_tic_input.sector], tic_id, cadence=120),
                                                 target_dir)
            lc = lcbuild.lc
            lc_df = pd.DataFrame(columns=['#time', 'flux', 'flux_err'])
            lc_df['#time'] = lc.time.value
            lc_df['flux'] = lc.flux.value
            lc_df['flux_err'] = lc.flux_err.value
            lc_df.to_csv(f"{target_dir}/{prepare_tic_input.sector}_{tic_id}_lc_short.csv")
            self.store_lc_data(lcbuild.lc_data, f"{target_dir}/{prepare_tic_input.sector}_{tic_id}_lcdata_long.csv")
        except Exception as e:
            logging.warning("Failed Short Cadence data for target " + str(prepare_tic_input.tic))
            logging.exception(e)
        logging.info("Trying to get long cadence info for " + str(prepare_tic_input.tic))
        try:
            lcbuild = \
                mission_lightcurve_builder.build(MissionObjectInfo([prepare_tic_input.sector], tic_id, cadence=1800),
                                                 target_dir)
            lc = lcbuild.lc
            lc_df = pd.DataFrame(columns=['#time', 'flux', 'flux_err'])
            lc_df['#time'] = lc.time.value
            lc_df['flux'] = lc.flux.value
            lc_df['flux_err'] = lc.flux_err.value
            lc_df.to_csv(f"{target_dir}/{prepare_tic_input.sector}_{tic_id}_lc_long.csv")
            self.store_lc_data(lcbuild.lc_data, f"{target_dir}/{prepare_tic_input.sector}_{tic_id}_lcdata_long.csv")
        except Exception as e:
            logging.warning("Failed Long Cadence data for target " + str(prepare_tic_input.tic))
            logging.exception(e)

    def prepare_negative_training_dataset(self, first_sector, cpus):
        logging.info("Preparing negatives")
        ois = OisManager(self.cache_dir).load_ois()
        inputs = []
        for sector in np.arange(first_sector, 27, 1):
            self.negative_dir = self.dir + "/original/"
            if not os.path.exists(self.negative_dir):
                os.mkdir(self.negative_dir)
            sector_url = self.SECTOR_URL.format(sector)
            logging.info("Preparing negatives for sector from url " + sector_url)
            sector_file = "sector.csv"
            tic_csv = open(sector_file, 'wb')
            request = requests.get(sector_url)
            tic_csv.write(request.content)
            tic_csv.close()
            tics_sector_df = pd.read_csv(sector_file, comment='#', sep=',')
            tics_sector_df = tics_sector_df.sample(n=500, random_state=0)
            os.remove(sector_file)
            tics_sector_df["TICID"] = "TIC " + tics_sector_df["TICID"].map(str)
            for tic in tics_sector_df["TICID"].unique():
                logging.info("Preparing negative input data for " + str(tic))
                if tic in ois["Object Id"]:
                    logging.warning(str(tic) + " has official candidates!")
                    continue
                inputs.append(PrepareTicInput(self.negative_dir, tic, ois, sector))
            with Pool(processes=cpus) as pool:
                pool.map(self.prepare_tic, inputs)

    # def prepare_tsfresh(self, positive_dir, negative_dir):
    #     tsfresh_short_df = pd.DataFrame(columns=['id', 'time', 'flux', 'flux_err', 'background_flux', 'quality', 'centroids_x',
    #                                        'centroids_y', 'motion_x', 'motion_y'])
    #     tsfresh_long_df = pd.DataFrame(columns=['id', 'time', 'flux', 'flux_err', 'background_flux', 'quality', 'centroids_x',
    #                                        'centroids_y', 'motion_x', 'motion_y'])
    #     tsfresh_tags_short = []
    #     tsfresh_tags_long = []
    #     for tic_dir in os.listdir(positive_dir):
    #         short_lc_dir = positive_dir + "/" + tic_dir + "/time_series_short.csv"
    #         if os.path.exists(short_lc_dir):
    #             lc_short_df = pd.read_csv(positive_dir + "/" + tic_dir + "/time_series_short.csv")
    #             lc_short_df['id'] = tic_dir
    #             tsfresh_short_df.append(lc_short_df)
    #             tsfresh_tags_short.append([tic_dir, 1])
    #         lc_long_df = pd.read_csv(positive_dir + "/" + tic_dir + "/time_series_long.csv")
    #         lc_long_df['id'] = tic_dir
    #         tsfresh_long_df.append(lc_long_df)
    #         tsfresh_tags_long.append([tic_dir, 1])
    #     for tic_dir in os.listdir(negative_dir):
    #         short_lc_dir = negative_dir + "/" + tic_dir + "/time_series_short.csv"
    #         if os.path.exists(short_lc_dir):
    #             lc_short_df = pd.read_csv(negative_dir + "/" + tic_dir + "/time_series_short.csv")
    #             lc_short_df['id'] = tic_dir
    #             tsfresh_short_df.append(lc_short_df)
    #             tsfresh_tags_short.append([tic_dir, 1])
    #         lc_long_df = pd.read_csv(negative_dir + "/" + tic_dir + "/time_series_long.csv")
    #         lc_long_df['id'] = tic_dir
    #         tsfresh_long_df.append(lc_long_df)
    #         tsfresh_tags_long.append([tic_dir, 0])
    #     tsfresh_tags_short = pd.Series(tsfresh_tags_short)
    #     tsfresh_tags_long = pd.Series(tsfresh_tags_long)
    #     # TODO tsfresh needs a dataframe with all the "time series" data (centroids, motion, flux, bck_flux...)
    #     # TODO with an id column specifying the target id and a "y" as a df containing the target ids and the classification
    #     # TODO tag. We need to check how to make this compatible with transit times tagging instead of entire curve
    #     # TODO classification. Maybe https://tsfresh.readthedocs.io/en/latest/text/forecasting.html is helpful.
    #     extracted_features_short = tsfresh.extract_relevant_features(tsfresh_short_df, tsfresh_tags_short, column_id='id',
    #                                                                  column_sort='time')
    #     extracted_features_long = tsfresh.extract_relevant_features(tsfresh_long_df, tsfresh_tags_long, column_id='id',
    #                                                                 column_sort='time')

    @staticmethod
    def prepare_cadence_set_dir(root_dir, mode="short"):
        """
        Creates the flattened directory for training and testing sets for the selected cadence method.
        @param root_dir: the directory where the data has been stored
        @param mode: the cadence mode [short|long]
        """
        flatten_dir = root_dir + "/" + mode
        if not os.path.exists(flatten_dir):
            os.mkdir(flatten_dir)
        positive_dir = root_dir + "/tp"
        false_positive_dir = root_dir + "/fp"
        negative_dir = root_dir + "/ntp"
        logging.info("Flattening positive dir %s", positive_dir)
        for file in os.listdir(positive_dir):
            src = positive_dir + "/" + file
            dst = flatten_dir + "/" + file + "_tp_"
            if os.path.exists(src + "/periodogram_" + mode + ".csv") and \
                    os.path.exists(src + "/time_series_" + mode + ".csv"):
                ois_df = pd.read_csv(src + "/ois.csv")
                for index, oi_row in ois_df.iterrows():
                    dst_index = dst + str(index)
                    os.symlink(src, dst_index)
        logging.info("Flattening negative dir %s", negative_dir)
        for file in os.listdir(false_positive_dir):
            src = false_positive_dir + "/" + file
            dst = flatten_dir + "/" + file + "_fp_"
            if os.path.exists(src + "/periodogram_" + mode + ".csv") and \
                    os.path.exists(src + "/time_series_" + mode + ".csv"):
                ois_df = pd.read_csv(src + "/ois.csv")
                for index, oi_row in ois_df.iterrows():
                    dst_index = dst + str(index)
                    os.symlink(src, dst_index)
        logging.info("Flattening false positive dir %s", false_positive_dir)
        np.random.seed(0)
        for file in os.listdir(negative_dir):
            tic_sector_file = negative_dir + "/" + file
            for file_in_sector in os.listdir(tic_sector_file):
                src = tic_sector_file + "/" + file_in_sector
                dst = flatten_dir + "/" + file_in_sector + "_ntp_0"
                if os.path.exists(src + "/periodogram_" + mode + ".csv") and \
                        os.path.exists(src + "/time_series_" + mode + ".csv"):
                    os.symlink(src, dst)

    @staticmethod
    def split_training_test_and_validation_sets(cadence_dir):
        labels = []
        filenames = []
        for file in os.listdir(cadence_dir):
            filenames.append(file)
            label = 0.0
            if "_tp_" in file:
                label = 1.0
            elif "_fp_" in file:
                label = 0.25 # TODO 0.25 or something else?
            labels.append(label)
        np.save(cadence_dir + '/filenames.npy', filenames)
        labels = keras.utils.to_categorical(labels)
        np.save(cadence_dir + '/labels.npy', labels)
        filenames_shuffled, y_labels_one_hot_shuffled = sklearn.utils.shuffle(filenames, labels)
        filenames_shuffled_numpy = np.array(filenames_shuffled)
        x_train_filenames, x_val_filenames, y_train, y_val = sklearn.model_selection.train_test_split(
            filenames_shuffled_numpy, labels, test_size=0.2, random_state=1)
        np.save(cadence_dir + '/x_train_filenames.npy', x_train_filenames)
        np.save(cadence_dir + '/y_train.npy', y_train)
        np.save(cadence_dir + '/x_val_filenames.npy', x_val_filenames)
        np.save(cadence_dir + '/y_val.npy', y_val)

cpus = 10
first_negative_sector = 1
training_data_dir = "/mnt/DATA-2/detrend/original/"
cache_dir = "/home/martin/"
ml_training_set_preparer = EmptyTargetsPreparer(training_data_dir, cache_dir)
ml_training_set_preparer.prepare_positive_training_dataset(cpus)
#ml_training_set_preparer.prepare_false_positive_training_dataset(cpus)
#ml_training_set_preparer.prepare_negative_training_dataset(first_negative_sector, cpus)
#MLSingleTransitsClassifier().load_candidate_single_transits(traid ning_data_dir, "tp")
#MlTrainingSetPreparer.prepare_cadence_set_dir(training_data_dir, "short")
#MLModelBuilder().get_model()
#MLModelBuilder()get_single_transit_model()


import os

import numpy as np
import pandas as pd
import tensorflow as tf
from sklearn.utils import shuffle
from tensorflow.keras.utils import Sequence

class SantoGenerator(Sequence):
    def __init__(self, lcs_dir, target_files, input_size=500, step_size=1, batch_size=500, shuffle=True, zero_epsilon=1e-7):
        self.lcs_dir = lcs_dir
        self.target_files = target_files
        self.input_size = input_size
        self.step_size = step_size
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.zero_epsilon = zero_epsilon
        self.steps_count, self.generator_info_df = self._count_inputs()

    def _count_inputs(self):
        batch_index = 0
        generator_info_df: pd.DataFrame = pd.DataFrame(columns=['filename', 'file_index', 'batch_index'])
        for target_file in self.target_files:
            flux = np.loadtxt(f'{self.lcs_dir}/{target_file}', delimiter=',')
            file_batch_indexes = np.arange(0, flux.shape[1], self.batch_size)
            for file_index in file_batch_indexes:
                generator_info_df = pd.concat([generator_info_df, pd.DataFrame.from_dict(
                                    {'filename': [target_file], 'file_index': [file_index],
                                     'batch_index': [batch_index]})], ignore_index=True)
                batch_index = batch_index + 1
        return batch_index - 1, generator_info_df

    def __len__(self):
        return self.generator_info_df['batch_index'].max()

    def __getitem__(self, index):
        filename: str = self.generator_info_df.iloc[index]['filename']
        file_index: int = self.generator_info_df.loc[index]['file_index']
        flux = np.loadtxt(f'{self.lcs_dir}/{filename}', delimiter=',')
        max_index = np.min([flux.shape[1] - self.input_size, file_index + self.batch_size])
        train_fluxes = np.full((self.batch_size, self.input_size), 0.5)
        train_tags = np.full((self.batch_size, 1), float(0))
        for iteration_index in np.arange(file_index, max_index):
            flux_data = flux[0][iteration_index:iteration_index + self.input_size] / 2
            train_fluxes[iteration_index - file_index] = flux_data
            tags_data_index = iteration_index + self.input_size // 2 - 1
            tags_data = np.array(flux[1][tags_data_index]).flatten()
            oot_mask = np.abs(tags_data - 1) < 1e-6
            tags_data[oot_mask] = float(0)
            tags_data[~oot_mask] = float(1)
            train_tags[iteration_index - file_index] = tags_data
        self.assert_in_range(filename, train_fluxes)
        np.all(train_tags != train_tags.astype(int)), "train_tags contains integer values!"
        return np.reshape(train_fluxes, (self.batch_size, 500, 1)), train_tags

    def assert_in_range(self, object_id, array, values_range=(0, 1)):
        if np.isnan(array).any():
            raise ValueError("Target " + str(object_id) + " contains NaN values")
        elif np.max(array) >= values_range[1]:
            raise ValueError("Target " + str(object_id) + " contains values > 1")
        elif np.min(array) <= values_range[0]:
            raise ValueError("Target " + str(object_id) + " contains values < 0")
        elif np.all(array == values_range[0]):
            raise ValueError("Target " + str(object_id) + " contains all values == 0")

    def on_epoch_end(self):
        if self.shuffle:
            self.generator_info_df = shuffle(self.generator_info_df)

    def class_weights(self):
        return {0: 1, 1: 1}

    def steps_per_epoch(self):
        return self.steps_count


import logging
import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import tensorflow as tf
from sklearn.utils import shuffle
from tensorflow.keras.utils import Sequence

class SantoGenerator(Sequence):
    def __init__(self, lcs_dir, target_files, input_size=20000, step_size=1, batch_size=500, shuffle=True,
                 zero_epsilon=1e-7, max_time_encoding=200, plot=False):
        self.lcs_dir = lcs_dir
        self.target_files = target_files
        self.input_size = input_size
        self.step_size = step_size
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.zero_epsilon = zero_epsilon
        self.steps_count, self.generator_info_df = self._count_inputs()
        self.max_time_encoding = max_time_encoding
        self.plot = plot

    def _count_inputs(self):
        batch_index = 0
        chunks_index = 0
        generator_info_df: pd.DataFrame = pd.DataFrame(columns=['filename', 'file_start_index', 'file_end_index', 'batch_index'])
        for target_file in self.target_files:
            filename = f'{self.lcs_dir}/{target_file}'
            flux = np.loadtxt(filename, delimiter=',')
            file_batch_indexes = np.arange(0, flux.shape[1], self.input_size // 2)
            stop = False
            for file_start_index in file_batch_indexes:
                if flux.shape[1] - file_start_index <= self.input_size:
                    file_start_index = flux.shape[1] - self.input_size - 1
                    if file_start_index < 0:
                        file_start_index = 0
                    stop = True
                file_end_index: int = file_start_index + self.input_size
                if file_end_index >= flux.shape[1]:
                    file_end_index = flux.shape[1] - 1
                    file_start_index = file_end_index - self.input_size
                if file_start_index < 0:
                    logging.error(
                        f"{filename} file_start_index should be greater than zero. Curve length is {file_end_index} - {file_start_index}")
                    continue
                if np.sum(flux[1, file_start_index:file_end_index] < 1e-7) / flux[1, file_start_index:file_end_index].shape[0] > 0.25:
                    continue

                # if np.all(np.abs(flux[2, file_start_index:file_end_index] - 1) < 1e-7):
                #     continue
                generator_info_df = pd.concat([generator_info_df, pd.DataFrame.from_dict(
                                    {'filename': [target_file], 'file_start_index': [file_start_index],
                                     'file_end_index': [file_end_index],
                                     'batch_index': [batch_index]})], ignore_index=True)
                chunks_index = chunks_index + 1
                if chunks_index % self.batch_size == 0:
                    batch_index = batch_index + 1
                if stop:
                    break
        generator_info_df.sort_values(['batch_index', 'filename', 'file_start_index'], ascending=True, inplace=True)
        return batch_index - 1, generator_info_df

    def __len__(self):
        return self.steps_count

    def create_tf_dataset(self):
        # Generator function to yield data in batches
        def dataset_generator():
            for batch_index in range(self.steps_count):
                batch_df = self.generator_info_df.loc[self.generator_info_df['batch_index'] == batch_index]
                train_data = np.zeros((self.batch_size, 3, self.input_size), dtype=np.float32)
                item_index = 0
                for row_index, batch_row in batch_df.iterrows():
                    filename = batch_row['filename']
                    flux = np.loadtxt(f'{self.lcs_dir}/{filename}', delimiter=',', dtype=np.float32)
                    file_start_index = batch_row['file_start_index']
                    file_end_index = batch_row['file_end_index']
                    train_data[item_index, 0] = flux[4, file_start_index:file_end_index]
                    train_data[item_index, 1] = flux[1, file_start_index:file_end_index] / 2
                    train_data[item_index, 2] = flux[3, file_start_index:file_end_index]
                    #print(np.any(flux[3, file_start_index:file_end_index] > 0.5))
                    yield (flux[[4, 1], file_start_index:file_end_index].swapaxes(0, 1),
                           flux[3, file_start_index:file_end_index])
                    item_index += 1
                train_data_swapped = train_data.swapaxes(1, 2)
                features = train_data_swapped[:, :, 0:2]
                labels = train_data_swapped[:, :, 2].reshape((self.batch_size, self.input_size))
                #yield features, labels
        dataset = tf.data.Dataset.from_generator(
            dataset_generator,
            output_signature=(
                tf.TensorSpec(shape=(self.input_size, 2), dtype=tf.float32),
                tf.TensorSpec(shape=(self.input_size), dtype=tf.float32)
            )
        )
        dataset = dataset.batch(self.batch_size).prefetch(tf.data.AUTOTUNE)
        return dataset

    def __getitem__(self, batch_index):
        batch_df = self.generator_info_df.loc[self.generator_info_df['batch_index'] == batch_index]
        train_data = np.zeros((self.batch_size, 3, self.input_size), dtype=np.float32)
        item_index = 0
        if len(batch_df) != self.batch_size:
            raise ValueError("More items than batch size")
        for row_index, batch_row in batch_df.iterrows():
            filename: str = batch_row['filename']
            flux = np.loadtxt(f'{self.lcs_dir}/{filename}', delimiter=',', dtype=np.float32)
            file_start_index: int = batch_row['file_start_index']
            file_end_index: int = batch_row['file_end_index']
            train_data[item_index, 0] = flux[4, file_start_index:file_end_index]
            train_data[item_index, 1] = flux[1, file_start_index:file_end_index] / 2
            train_data[item_index, 2] = flux[3, file_start_index:file_end_index]
            # self.assert_tags_value(filename, train_data[item_index, :, 2])
            # self.assert_in_range(filename, train_data[item_index, :])
            if self.plot:
                fig, axs = plt.subplots(2, 1, figsize=(15, 10))
                axs[0].scatter(flux[0, file_start_index:file_end_index], train_data[item_index, 1])
                axs[1].plot(flux[0, file_start_index:file_end_index], train_data[item_index, 2])
                plt.show()
            item_index = item_index + 1
        # np.all(train_data[item_index, :, 2] != train_data[item_index, :, 2].astype(int)), "train_tags contains integer values!"
        # if np.any(np.abs(train_data[item_index, :, 2]) - 1 < 1e-6):
        #     print("Train tags greater than one")
        train_data_swapped = train_data.swapaxes(1, 2)
        return train_data_swapped[:, :, 0:2], train_data_swapped[:, :, 2].reshape((self.batch_size, self.input_size))

    def assert_in_range(self, object_id, array, values_range=(0, 1)):
        if np.isnan(array).any():
            raise ValueError("Target " + str(object_id) + " contains NaN values")
        elif np.max(array) > values_range[1]:
            raise ValueError("Target " + str(object_id) + " contains values > 1")
        elif np.min(array) < values_range[0]:
            raise ValueError("Target " + str(object_id) + " contains values < 0")

    def assert_tags_value(self, object_id, array, values_range=(0, 1)):
        if np.any((array > values_range[0]) & (array < values_range[1])):
            raise ValueError("Target " + str(object_id) + " contains invalid values")

    def on_epoch_end(self):
        if self.shuffle:
            self.generator_info_df = shuffle(self.generator_info_df)

    def class_weights(self):
        return {0: 1, 1: 1}

    def steps_per_epoch(self):
        return self.steps_count

#
# class SantoGenerator2:
#     def __init__(self, lcs_dir, target_files, input_size=20000, batch_size=64, shuffle=True):
#         self.lcs_dir = lcs_dir
#         self.target_files = target_files
#         self.input_size = input_size
#         self.batch_size = batch_size
#         self.shuffle = shuffle
#         self.file_indices = self._prepare_indices()
#
#     def _prepare_indices(self):
#         indices = []
#         for file in self.target_files:
#             flux = np.loadtxt(f'{self.lcs_dir}/{file}', delimiter=',', dtype=np.float32)
#             file_length = flux.shape[1]
#             chunks = np.arange(0, file_length, self.input_size // 2)
#             for start_index in chunks:
#                 end_index = min(start_index + self.input_size, file_length)
#                 if end_index - start_index == self.input_size:
#                     indices.append((file, start_index, end_index))
#         if self.shuffle:
#             np.random.shuffle(indices)
#         return indices
#
#     def _process_file(self, file, start_index, end_index):
#         flux = np.loadtxt(f'{self.lcs_dir}/{file}', delimiter=',', dtype=np.float32)
#         time_data = np.sin((2 * np.pi / 200) * flux[0, start_index:end_index])
#         time_data = np.clip(time_data, 1e-7, 1 - 1e-7)
#         flux_data = flux[1, start_index:end_index] / 2
#         tags_data = flux[2, start_index:end_index]
#         tags_data = np.where((tags_data < 1e-5) | (np.abs(tags_data - 1) < 1e-7), 0.0, 1.0).astype(np.float32)
#         train_data = np.stack([time_data, flux_data], axis=-1)
#         return train_data, tags_data
#
#     def generator(self):
#         """Python generator function that yields batches."""
#         while True:  # Loop forever so Keras can run multiple epochs
#             batch_data = []
#             batch_labels = []
#             for file, start_index, end_index in self.file_indices:
#                 data, labels = self._process_file(file, start_index, end_index)
#                 batch_data.append(data)
#                 batch_labels.append(labels)
#                 if len(batch_data) == self.batch_size:
#                     yield np.array(batch_data, dtype=np.float32), np.array(batch_labels, dtype=np.float32)
#                     batch_data = []
#                     batch_labels = []


from numpy.random import default_rng

from exoml.ete6.ete6_generator import Ete6ModelGenerator
from exoml.ml.encoding.time_position import value_encode_times
from sklearn.utils import shuffle
import numpy as np
import pandas as pd


class TranspotModelGenerator(Ete6ModelGenerator):
    def __init__(self, injected_objects_df, lcs_dir, batch_size, input_sizes, zero_epsilon=1e-7,
                 measurements_per_point=13):
        super().__init__()
        self.injected_objects_df = injected_objects_df
        self.lcs_dir = lcs_dir
        self.batch_size = batch_size
        self.input_sizes = input_sizes
        self.zero_epsilon = zero_epsilon
        self.random_number_generator = default_rng()
        self.measurements_per_point = measurements_per_point

    def __len__(self):
        return (np.ceil(len(self.injected_objects_df) / float(self.batch_size))).astype(int)

    def class_weights(self):
        return {0: 1, 1: 1}

    def __getitem__(self, idx):
        max_index = (idx + 1) * self.batch_size
        max_index = len(self.injected_objects_df) if max_index > len(self.injected_objects_df) else max_index
        target_indexes = np.arange(idx * self.batch_size, max_index, 1)
        target_indexes = shuffle(target_indexes)
        injected_objects_df = self.injected_objects_df.iloc[target_indexes]
        star_array = np.empty((len(target_indexes), 11, 1))
        global_flux_array = np.empty((len(target_indexes), self.input_sizes[0], self.measurements_per_point))
        batch_data_values = np.empty((len(target_indexes), 1))
        i = 0
        for df_index, target_row in injected_objects_df.iterrows():
            target_id = int(target_row['TIC ID'])
            leading_zeros_object_id = '{:09}'.format(target_id)
            lc_filename = str(leading_zeros_object_id) + '_lc.csv'
            star_filename = str(leading_zeros_object_id) + '_star.csv'
            star_df = pd.read_csv(self.lcs_dir + '/' + star_filename, index_col=False)
            star_df = self._prepare_input_star(star_df)
            lc_df = pd.read_csv(self.lcs_dir + '/' + lc_filename, usecols=['#time', 'flux', 'flux_0', 'flux_1',
                                                                           'flux_2', 'flux_3', 'flux_4',
                                                                           'flux_err', 'centroid_x',
                                                                           'centroid_y', 'motion_x', 'motion_y',
                                                                           'bck_flux'], low_memory=True)
            lc_df = self._prepare_input_lc(lc_df)
            type = target_row['type']
            batch_data_values[i] = 0 if type == 'none' else 1
            # not_null_times_args = np.argwhere(lc_df['#time'].to_numpy() > 0).flatten()
            # lc_df = lc_df.iloc[not_null_times_args]
            star_array[i] = np.transpose([star_df.to_numpy()])
            global_flux_array[i] = lc_df.to_numpy()
            assert not np.isnan(star_array[i]).any() and not np.isinf(star_array[i]).any()
            assert not np.isnan(global_flux_array[i]).any() and not np.isinf(global_flux_array[i]).any()
            if np.max(global_flux_array[i]) >= 2:
                print("MAX FLUX VALUE OVERCAME BY " + str(target_id))
            assert np.max(global_flux_array[i]) < 2
            i = i + 1
        return [star_array, global_flux_array], batch_data_values

    def _prepare_input_lc(self, input_df):
        result_df = super()._prepare_input_lc(input_df)
        args_zero = np.argwhere(result_df["#time"].to_numpy() <= self.zero_epsilon).flatten()
        first_non_zero_arg = np.argwhere(result_df["#time"].to_numpy() > self.zero_epsilon).flatten()[0]
        time = result_df["#time"].to_numpy()
        result_df["#time"] = value_encode_times(time - time[first_non_zero_arg], max_time=100)
        result_df["#time"].iloc[args_zero] = self.zero_epsilon
        return result_df

# df = pd.read_csv("/mnt/DATA-2/ete6/injected_objects.csv", index_col=False)
# df = shuffle(df)
# df.reset_index(inplace=True, drop=True)
# img = IatsonModelGenerator(df, "/mnt/DATA-2/ete6/lcs", 20, [11, 2500, 500, 500, 500, 500, 500, 500])
# img.__getitem__(20)
